__author__ = 'Robert Meyer'

import numpy as np
import sys
import unittest

from pypet.parameter import Parameter, PickleParameter, ArrayParameter,\
    SparseParameter, ObjectTable, Result, SparseResult, PickleResult, BaseParameter
from pypet.trajectory import Trajectory
import pickle
import scipy.sparse as spsp
import pypet.pypetexceptions as pex
import warnings
import pandas as pd
import pypet.utils.comparisons as comp
from pypet.utils.helpful_classes import ChainMap
from pypet.utils.explore import cartesian_product
import pypet.pypetconstants as pypetconstants
from pypet.tests.testutils.ioutils import run_suite, parse_args, make_temp_dir
from pypet.tests.testutils.data import TrajectoryComparator


class ParameterTest(TrajectoryComparator):

    tags = 'unittest', 'parameter'

    def test_type_error_for_not_supported_data(self):

        for param in self.param.values():
            if not isinstance(param, PickleParameter):
                with self.assertRaises(TypeError):
                    param._values_of_same_type(ChainMap(),ChainMap())

                with self.assertRaises(TypeError):
                    param._equal_values(ChainMap(),ChainMap())

    def test_store_load_with_hdf5(self):
        traj_name = 'test_%s' % self.__class__.__name__
        filename = make_temp_dir(traj_name + '.hdf5')
        traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                          filename = filename, overwrite_file=True)

        for param in self.param.values():
            traj.f_add_parameter(param)

        traj.f_store()

        new_traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                              filename = filename)

        new_traj.f_load(load_data=2)
        self.compare_trajectories(traj, new_traj)

    def test_store_load_with_hdf5_no_data(self):
        traj_name = 'test_%s' % self.__class__.__name__
        filename = make_temp_dir(traj_name + 'nodata.hdf5')
        traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                          filename = filename, overwrite_file=True)

        for param in self.param.values():
            param._data = None
            traj.f_add_parameter(param)

        traj.f_store()

        new_traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                              filename = filename)

        new_traj.f_load(load_data=2)
        self.compare_trajectories(traj, new_traj)

    def test_type_error_for_exploring_if_range_does_not_match(self):

        param = self.param['val1']

        with self.assertRaises(TypeError):
            param._explore(['a','b'])

        with self.assertRaises(TypeError):
            param._explore([ChainMap(),ChainMap()])

        with self.assertRaises(ValueError):
            param._explore([])

    def test_cannot_expand_and_not_explore_throwing_type_error(self):

        for param in self.param.values():
            if not param.f_has_range():
                with self.assertRaises(TypeError):
                    param._expand([12,33])
            else:
                with self.assertRaises(TypeError):
                    param._explore([12,33])


    def test_equal_values(self):
        for param in self.param.values():
            self.assertTrue(param._equal_values(param.f_get(), param.f_get()))

            self.assertFalse(param._equal_values(param.f_get(),23432432432))

            self.assertFalse(param._equal_values(param.f_get(),ChainMap()))

            if not isinstance(param, PickleParameter):
                with self.assertRaises(TypeError):
                     self.assertFalse(param._equal_values(ChainMap(),ChainMap()))


    def test_parameter_locking(self):
        for param in self.param.values():

            if not param.v_explored:
                self.assertFalse(param.v_locked, 'Param %s is locked' % param.v_full_name)
                param.f_lock()
            else:
                self.assertTrue(param.v_locked, 'Param %s is locked' % param.v_full_name)

            with self.assertRaises(pex.ParameterLockedException):
                param.f_set(3)

            with self.assertRaises(pex.ParameterLockedException):
                param._explore([3])

            with self.assertRaises(pex.ParameterLockedException):
                param._expand([3])

            with self.assertRaises(pex.ParameterLockedException):
                param._shrink()

            with self.assertRaises(pex.ParameterLockedException):
                param.f_empty()

    def test_param_accepts_not_unsupported_data(self):

        for param in self.param.values():
            if not isinstance(param, PickleParameter):
                with self.assertRaises(TypeError):
                    param.f_set(ChainMap())

    def test_parameter_access_throws_ValueError(self):

        for name,param in self.param.items():
            if name in self.explore_dict:
                self.assertTrue(param.f_has_range())

                with self.assertRaises(ValueError):
                    param._set_parameter_access(1232121321321)

    def test_values_of_same_type(self):
        for param in self.param.values():
            self.assertTrue(param._values_of_same_type(param.f_get(),param.f_get()))

            if not isinstance(param.f_get(), int):
                self.assertFalse(param._values_of_same_type(param.f_get(),23432432432))

            self.assertFalse(param._values_of_same_type(param.f_get(),ChainMap()))

            if not isinstance(param, PickleParameter):
                with self.assertRaises(TypeError):
                     self.assertFalse(param._equal_values(ChainMap(),ChainMap()))

    def test_meta_settings(self):
        for key, param in self.param.items():
            self.assertEqual(param.v_full_name, self.location+'.'+key)
            self.assertEqual(param.v_name, key)
            self.assertEqual(param.v_location, self.location)


    def make_params(self):
        self.param = {}
        for key, val in self.data.items():
            self.param[key] = Parameter(self.location+'.'+key, val, comment=key)


    def setUp(self):

        if not hasattr(self,'data'):
            self.data={}

        self.data['val0']= 1
        self.data['val1']= 1.0
        self.data['val2']= True
        self.data['val3'] = 'String'
        self.data['npfloat'] = np.array([1.0,2.0,3.0])
        self.data['npfloat_2d'] = np.array([[1.0,2.0],[3.0,4.0]])
        self.data['npbool']= np.array([True,False, True])
        self.data['npstr'] = np.array(['Uno', 'Dos', 'Tres'])
        self.data['npint'] = np.array([1,2,3])
        self.data['alist'] = [1,2,3,4]
        self.data['atuple'] = (1,2,3,4)

        self.location = 'MyName.Is.myParam'

        self.dynamic_imports = []


        self.make_params()


        # with self.assertRaises(AttributeError):
        #     self.param.val0.f_set([[1,2,3],[1,2,3]])

        #Add explortation:
        self.explore()

    def test_get_item(self):
        for paramname in self.explore_dict:
            param = self.param[paramname]
            val1=param.f_get_range()[1]
            val2=param[1]
            self.assertTrue(comp.nested_equal(val1,val2), '%s != %s' % (str(val1),str(val2)))

    def test_get_data(self):
        for paramname in self.param:
            param = self.param[paramname]
            val1=param.data
            val2=param.f_get()
            self.assertTrue(comp.nested_equal(val1,val2), '%s != %s' % (str(val1),str(val2)))
            val3=param['data']
            self.assertTrue(comp.nested_equal(val3,val2), '%s != %s' % (str(val3),str(val2)))

            val1=param.default
            val2=param._default
            self.assertTrue(comp.nested_equal(val1,val2), '%s != %s' % (str(val1),str(val2)))
            val3=param['default']
            self.assertTrue(comp.nested_equal(val3,val2), '%s != %s' % (str(val3),str(val2)))
            val4=param.f_get_default()
            self.assertTrue(comp.nested_equal(val4,val2), '%s != %s' % (str(val4),str(val2)))
            val5 = param[-1]
            self.assertTrue(comp.nested_equal(val5,val2), '%s != %s' % (str(val5),str(val2)))

    def test_type_error_for_get_item(self):
        for name,param in self.param.items():
            if not name in self.explore_dict:
                with self.assertRaises(TypeError):
                    param[1]

    def test_type_error_for_shrink(self):
        for name, param in self.param.items():
            if not name in self.explore_dict:
                with self.assertRaises(TypeError):
                    param._shrink()

    def explore(self):
        self.explore_dict=cartesian_product({'npstr':[np.array(['Uno', 'Dos', 'Tres']),
                               np.array(['Cinco', 'Seis', 'Siette']),
                            np.array(['Ocho', 'Nueve', 'Diez'])],
                           'val0':[1,2,3], 'alist':[[1,2,3,4],[3,4,5,6]]})

        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            old_data = self.param[key]._data
            self.param[key]._data = None
            with self.assertRaises(TypeError):
                self.param[key]._explore(vallist)
            self.param[key]._data = old_data
            self.param[key]._explore(vallist)

    def test_the_insertion_made_implicetly_in_setUp(self):
        for key, val in self.data.items():

            if not key in self.explore_dict:
                self.param[key]._restore_default()
                param_val = self.param[key].f_get()
                self.assertTrue(np.all(repr(val) == repr(param_val)),'%s != %s'  %(str(val),str(param_val)))



    def test_expanding_type_error(self):
        for name,param in self.param.items():
            if not name in self.explore_dict:
                #Test locking
                with self.assertRaises(TypeError):
                    param._expand([1,2,3])

                #Test wron param type
                with self.assertRaises(TypeError):
                    param.f_unlock()
                    param._expand([1,2,3])


    def test_rename(self):
        for name,param in self.param.items():
            param._rename('test.test.wirsing')
            self.assertTrue(param.v_name=='wirsing')
            self.assertTrue(param.v_full_name=='test.test.wirsing')
            self.assertTrue(param.v_location=='test.test')



    def test_expanding(self):
        for name,param in self.param.items():
            if name in self.explore_dict:
                param.f_unlock()
                param._expand(self.explore_dict[name])

                self.assertTrue(len(param) == 2*len(self.explore_dict[name]),
                                'Expanding of %s did not work.' % name)


    def test_exploration(self):
        for key, vallist in self.explore_dict.items():

            param = self.param[key]

            for idx, val in enumerate(vallist):
                assert isinstance(param, BaseParameter)
                param._set_parameter_access(idx)

                self.assertTrue(np.all(repr(param.f_get())==repr(val))),'%s != %s'%( str(param.f_get()),str(val))

                param_val = self.param[key].f_get_range()[idx]
                self.assertTrue(np.all(str(val) == str(param_val)),'%s != %s'  %(str(val),str(param_val)))

            param._restore_default()
            self.assertTrue(param.v_explored and param.f_has_range(), 'Error for %s' % key)
            val = self.data[key]
            self.assertTrue(np.all(repr(param.f_get())==repr(val))),'%s != %s'%( str(param.f_get()),str(val))


    def test_storage_and_loading(self):

        for key, param in self.param.items():
            store_dict = param._store()

            # Due to smart storing the storage dict should be small and only contain 5 items or less
            # 1 for data, 1 for reference, and 3 for the array/matrices/items
            if param.f_has_range():
                if isinstance(param,(ArrayParameter, PickleParameter)) and \
                        not isinstance(param, SparseParameter):
                    self.assertTrue(len(store_dict)<7)
                # For sparse parameter it is more:
                if isinstance(param, SparseParameter):
                    self.assertTrue(len(store_dict)<23)



            constructor = param.__class__

            param.f_unlock()
            param.f_empty()

            param = constructor('', 42)

            param._load(store_dict)

            param._rename(self.location+'.'+key)

            self.param[key] = param


        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_exploration()

        self.test_meta_settings()


    def test_pickling_without_multiprocessing(self):
        for key, param in self.param.items():
            param.f_unlock()
            param.v_full_copy=True

            dump = pickle.dumps(param)

            newParam = pickle.loads(dump)



            self.param[key] = newParam

        self.test_exploration()

        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_meta_settings()


    def test_pickling_with_mocking_multiprocessing(self):

        self.test_exploration()

        for key, param in self.param.items():
            param.f_unlock()
            param.v_full_copy=False

            dump = pickle.dumps(param)

            newParam = pickle.loads(dump)



            self.param[key] = newParam

            if key in self.explore_dict:
                self.assertTrue(not newParam.f_has_range() and newParam.v_explored)

        #self.test_exploration()

        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_meta_settings()



    def test_resizing_and_deletion(self):

        for key, param in self.param.items():
            param.f_lock()
            with self.assertRaises(pex.ParameterLockedException):
                 param.f_set(42)

            with self.assertRaises(pex.ParameterLockedException):
                param._shrink()


            param.f_unlock()


            if len(param)> 1:
                self.assertTrue(param.f_has_range())

            if param.f_has_range():
                self.assertTrue(len(param)>1)
                param._shrink()

            self.assertTrue(len(param) == 1)

            self.assertFalse(param.f_is_empty())
            self.assertFalse(param.f_has_range())



            param.f_empty()

            self.assertTrue(param.f_is_empty())
            self.assertFalse(param.f_has_range())

class ArrayParameterTest(ParameterTest):

    tags = 'unittest', 'parameter', 'array'

    def setUp(self):


        if not hasattr(self,'data'):
            self.data= {}

        self.data['myemptytuple'] = ()
        self.data['myemptylist'] = []

        self.data['myinttuple'] = (1,2,3)
        self.data['mydoubletuple'] = (42.0,43.7,33.3)
        self.data['mystringtuple'] = ('Eins','zwei','dr3i')

        super(ArrayParameterTest,self).setUp()


    def make_params(self):
        self.param = {}
        for key, val in self.data.items():
            self.param[key] = ArrayParameter(self.location+'.'+key, val, comment=key)


    def explore(self):

        matrices = []


        self.explore_dict=cartesian_product({'npstr':[np.array(['Uno', 'Dos', 'Tres']),
                               np.array(['Cinco', 'Seis', 'Siette']),
                            np.array(['Ocho', 'Nueve', 'Diez'])],
                           'val0':[1,2,3],
                           'myinttuple':[(1,2,1),(4,5,6),(5,6,7)],
                            'alist':[[1,2,3,4],['wooot']]} ,
                                            (('npstr','val0'),'myinttuple', 'alist'))

        ### Convert the explored stuff into numpy arrays
        #for idx, val in enumerate(self.explore_dict['myinttuple']):
         #   self.explore_dict['myinttuple'][idx] = np.array(val)


        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param[key]._explore(vallist)


    def test_store_load_with_hdf5(self):
        return super(ArrayParameterTest, self).test_store_load_with_hdf5()




class PickleParameterTest(ParameterTest):

    tags = 'unittest', 'parameter', 'pickle'

    def setUp(self):

        if not hasattr(self,'data'):
            self.data={}

        self.data['spsparse_csc'] = spsp.lil_matrix((1000,100))
        self.data['spsparse_csc'][1,2] = 44.5
        self.data['spsparse_csc'] = self.data['spsparse_csc'].tocsc()

        self.data['spsparse_csr'] = spsp.lil_matrix((2222,22))
        self.data['spsparse_csr'][1,3] = 44.5
        self.data['spsparse_csr'] = self.data['spsparse_csr'].tocsr()

        self.data['spsparse_lil'] = spsp.lil_matrix((111,111))
        self.data['spsparse_lil'][3,2] = 44.5

        super(PickleParameterTest,self).setUp()


    def make_params(self):
        self.param = {}
        count = 0
        self.protocols={}
        for key, val in self.data.items():
            self.param[key] = PickleParameter(self.location+'.'+key, val,
                                              comment=key, protocol=count)
            self.protocols[key]=count
            count +=1
            count = count % 3


    def test_meta_settings(self):
        for key, param in self.param.items():
            self.assertEqual(param.v_full_name, self.location+'.'+key)
            self.assertEqual(param.v_name, key)
            self.assertEqual(param.v_location, self.location)
            self.assertEqual(param.v_protocol, self.protocols[key], '%d != %d' %
                                                        (param.v_protocol, self.protocols[key]))

    def explore(self):

        matrices = []


        for irun in range(3):

            spsparse_lil = spsp.lil_matrix((111,111))
            spsparse_lil[3,2] = 44.5*irun

            matrices.append(spsparse_lil)


        self.explore_dict=cartesian_product({'npstr':[np.array(['Uno', 'Dos', 'Tres']),
                               np.array(['Cinco', 'Seis', 'Siette']),
                            np.array(['Ocho', 'Nueve', 'Diez'])],
                           'val0':[1,2,3],
                           'spsparse_lil' : matrices}, (('npstr','val0'),'spsparse_lil'))




        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param[key]._explore(vallist)


class SparseParameterTest(ParameterTest):

    tags = 'unittest', 'parameter', 'sparse'

    def setUp(self):

        if not hasattr(self,'data'):
            self.data={}

        self.data['spsparse_csc'] = spsp.lil_matrix((1000,100))
        self.data['spsparse_csc'][1,2] = 44.5
        self.data['spsparse_csc'] = self.data['spsparse_csc'].tocsc()

        self.data['spsparse_csr'] = spsp.lil_matrix((2222,22))
        self.data['spsparse_csr'][1,3] = 44.5
        self.data['spsparse_csr'] = self.data['spsparse_csr'].tocsr()

        self.data['spsparse_bsr'] = spsp.lil_matrix((111,111))
        self.data['spsparse_bsr'][3,2] = 44.5
        self.data['spsparse_bsr'] = self.data['spsparse_bsr'].tocsr().tobsr()

        self.data['spsparse_dia'] = spsp.lil_matrix((111,111))
        self.data['spsparse_dia'][3,2] = 44.5
        self.data['spsparse_dia'] = self.data['spsparse_dia'].tocsr().todia()

        super(SparseParameterTest,self).setUp()


    def make_params(self):
        self.param = {}
        for key, val in self.data.items():
            self.param[key] = SparseParameter(self.location+'.'+key, val, comment=key)

    def explore(self):

        matrices_csr = []
        for irun in range(3):

            spsparse_csr = spsp.csr_matrix((111,111))
            spsparse_csr[3,2+irun] = 44.5*irun

            matrices_csr.append(spsparse_csr)

        matrices_csc = []
        for irun in range(3):

            spsparse_csc = spsp.csc_matrix((111,111))
            spsparse_csc[3,2+irun] = 44.5*irun

            matrices_csc.append(spsparse_csc)

        matrices_bsr = []
        for irun in range(3):

            spsparse_bsr = spsp.csr_matrix((111,111))
            spsparse_bsr[3,2+irun] = 44.5*irun

            matrices_bsr.append(spsparse_bsr.tobsr())

        matrices_dia = []
        for irun in range(3):

            spsparse_dia = spsp.csr_matrix((111,111))
            spsparse_dia[3,2+irun] = 44.5*irun

            matrices_dia.append(spsparse_dia.todia())


        self.explore_dict=cartesian_product({'npstr':[np.array(['Uno', 'Dos', 'Tres']),
                               np.array(['Cinco', 'Seis', 'Siette']),
                            np.array(['Ocho', 'Nueve', 'Diez'])],
                           'val0':[1,2,3],
                           'spsparse_csr' : matrices_csr,
                           'spsparse_csc' : matrices_csc,
                           'spsparse_bsr' : matrices_bsr,
                           'spsparse_dia' : matrices_dia},
                            (('npstr','val0'),('spsparse_csr',
                            'spsparse_csc', 'spsparse_bsr','spsparse_dia')) )




        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param[key]._explore(vallist)


class ResultTest(TrajectoryComparator):

    tags = 'unittest', 'result'

    def make_results(self):
        self.results= {}
        self.results['test.res.on_constructor']=self.Constructor('test.res.on_constructor',
                                                                 **self.data)
        self.results['test.res.args']=self.Constructor('test.res.args')
        self.results['test.res.kwargs']=self.Constructor('test.res.kwargs')
        self.results['test.res.setitem']=self.Constructor('test.res.setitem')

        self.results['test.res.args'].f_set(*list(self.data.values()))
        self.results['test.res.kwargs'].f_set(**self.data)

        for key, value in self.data.items():
            self.results['test.res.setitem'][key]=value

    def test_set_item_via_number(self):
        res = self.results['test.res.args']

        res[0] = 'Hi'
        res[777] = 777

        self.assertTrue(getattr(res, res.v_name) == 'Hi')
        self.assertTrue(res.f_get(0) == 'Hi')
        self.assertTrue(getattr(res, res.v_name + '_777') == 777)
        self.assertTrue(res[777] == 777)
        self.assertTrue(res.f_get(777) == 777)

        self.assertTrue(0 in res)
        self.assertTrue(777 in res)
        self.assertTrue(99999999 not in res)

        del res[0]
        self.assertTrue(0 not in res)

        del res[777]
        self.assertTrue(777 not in res)

    def test_iter(self):
        for res in self.results.values():
            keyset1 = set([x for x in res])
            keyset2 = set(res.f_to_dict().keys())
            self.assertTrue(keyset1 == keyset2)


    def make_constructor(self):
        self.Constructor=Result
        self.dynamic_imports = [Result]

    def test_warning(self):
        for res in self.results.values():
            with warnings.catch_warnings(record=True) as w:
                res._stored=True
                res.XxxXXXXxxxx = 14

    def test_f_get_many_items(self):
        for res in self.results.values():
            if 'integer' in res and 'float' in res:
                myreslist = res.f_get('integer', 'float')
                self.assertEqual([self.data['integer'], self.data['float']],myreslist)

    def test_getattr_and_setattr(self):
        for res in self.results.values():
            res.iamanewvar = 42
            self.assertTrue(res.iamanewvar,42)
            with self.assertRaises(AttributeError):
                res.iamanonexistingvar

    def test_deletion_throws_error_if_item_not_there(self):
        for res in self.results.values():
            with self.assertRaises(AttributeError):
                del res.idonotexistforsure


    def test_get_data(self):
        for res in self.results.values():
            if len(res._data)==1:
                val1=res.data
                val2=res.f_get()
                self.assertTrue(comp.nested_equal(val1,val2), '%s != %s' % (str(val1),str(val2)))
                val3=res['data']
                self.assertTrue(comp.nested_equal(val3,val2), '%s != %s' % (str(val3),str(val2)))
            else:
                with self.assertRaises(AttributeError):
                    res.data
                with self.assertRaises(AttributeError):
                    res['data']
                with self.assertRaises(ValueError):
                    res.f_get()

    def test_f_get_errors(self):

        res = Result('test')

        with self.assertRaises(AttributeError):
            res.f_get()

        res.f_set(1,2,42)

        with self.assertRaises(ValueError):
            res.f_get()

    def test_contains(self):
        if 'test.res.kwargs' in self.results:
            res = self.results['test.res.kwargs']
            self.assertTrue('integer' in res)

            self.assertFalse('integoAr' in res)

    def test_deletion(self):

        for res in self.results.values():
            for key in res.f_to_dict():
                delattr(res,key)

            self.assertTrue(res.f_is_empty())

    def test_set_numbering(self):
        int_list = list(range(10))
        for res in self.results.values():
            res.f_set(*int_list)

            self.assertEqual(res.f_get(*int_list), int_list)

    def test_dir(self):
        res = self.results['test.res.on_constructor']

        dirlist = dir(res)

        self.assertTrue('integer' in dirlist)
        self.assertTrue('dict' in dirlist)
        self.assertTrue('float' in dirlist)
        self.assertTrue('tuple' in dirlist)

        res = self.results['test.res.args']

        dirlist = dir(res)

        self.assertTrue(res.v_name in dirlist)
        self.assertTrue('%s_1' % res.v_name in dirlist)

    def setUp(self):

        if not hasattr(self,'data'):
            self.data={}

        self.data['integer'] = 42
        self.data['float'] = 42.424242
        self.data['string'] = 'TestString! 66'
        self.data['long'] = 444444444444444444
        self.data['numpy_array'] = np.array([[3232.3,323232323232.32323232],[4.,4.]])
        self.data['tuple'] = (444,444,443)
        self.data['list'] = ['3','4','666']
        self.data['dict'] = {'a':'b','c':42, 'd': (1,2,3)}
        self.data['object_table'] = ObjectTable(data={'characters':['Luke', 'Han', 'Spock'],
                                    'Random_Values' :[42,43,44],
                                    'Arrays': [np.array([1,2]),np.array([3, 4]), np.array([5,5])]})
        self.data['pandas_frame'] = pd.DataFrame(data={'characters':['Luke', 'Han', 'Spock'],
                                    'Random_Values' :[42,43,44],
                                    'Doubles': [1.2,3.4,5.6]})

        self.data['nested_data.hui.buh.integer'] = 42

        myframe = pd.DataFrame(data ={'TC1':[1,2,3],'TC2':['Waaa',np.nan,''],'TC3':[1.2,42.2,np.nan]})


        myseries = myframe['TC1']

        self.data['series'] = myseries

        self.make_constructor()
        self.make_results()

    def test_store_load_with_hdf5(self):
        traj_name = 'test_%s' % self.__class__.__name__
        filename = make_temp_dir(traj_name + '.hdf5')
        traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                          filename = filename, overwrite_file=True)

        for res in self.results.values():
            traj.f_add_result(res)

        traj.f_store()

        new_traj = Trajectory(name=traj_name, dynamic_imports=self.dynamic_imports,
                              filename = filename)

        new_traj.f_load(load_data=2)
        self.compare_trajectories(traj, new_traj)

    def test_rename(self):
        for name,res in self.results.items():
            res._rename('test.test.wirsing')
            self.assertTrue(res.v_name=='wirsing')
            self.assertTrue(res.v_full_name=='test.test.wirsing')
            self.assertTrue(res.v_location=='test.test')

    def test_emptying(self):
        for res in self.results.values():

            self.assertFalse(res.f_is_empty())
            res.f_empty()

            self.assertTrue(res.f_is_empty())


    def test_string_representation(self):
        self.results['kkk']=self.Constructor('test.res.kkk', answer=42)
        self.results['rrr']=self.Constructor('test.res.rrr')
        for res in self.results.values():
            resstrlist = []
            strlen = 0
            for key in res._data:
                val = res._data[key]
                resstr = '%s=%s, ' % (key, repr(val))
                resstrlist.append(resstr)

                strlen += len(resstr)
                if strlen > pypetconstants.HDF5_STRCOL_MAX_VALUE_LENGTH:
                    break

            return_string = "".join(resstrlist)
            if len(return_string) > pypetconstants.HDF5_STRCOL_MAX_VALUE_LENGTH:
                return_string =\
                    return_string[0:pypetconstants.HDF5_STRCOL_MAX_VALUE_LENGTH - 3] + '...'
            else:
                return_string = return_string[0:-2] # Delete the last `, `


            valstr = res.f_val_to_str()
            self.assertTrue(return_string==valstr)
            self.assertTrue(len(valstr)<=pypetconstants.HDF5_STRCOL_MAX_VALUE_LENGTH)


    def test_meta_settings(self):
        for key, res in self.results.items():
            self.assertEqual(res.v_full_name, key)
            self.assertEqual(res.v_name, key.split('.')[-1])
            self.assertEqual(res.v_location, '.'.join(key.split('.')[0:-1]))

    def test_natural_naming(self):
        for res_name,res in self.results.items():
            for key, val1 in res.f_to_dict().items():
                val2 = getattr(res, key)
                self.assertTrue(comp.nested_equal(val1,val2))

    def test_get_item(self):
        for res_name,res in self.results.items():
            for key, val1 in res.f_to_dict().items():
                val2 = res[key]
                self.assertTrue(comp.nested_equal(val1,val2))

    def test_f_to_dict_no_copy(self):
        for res_name,res in self.results.items():
            for key, val1 in res.f_to_dict(copy=False).items():
                val2 = res[key]
                self.assertTrue(comp.nested_equal(val1,val2))


    def test_Attribute_error_for_get_item(self):
        for res in self.results.values():
            with self.assertRaises(AttributeError):
                res['IDONOTEXIST']

    def test_reject_outer_data_structure(self):
        for res in self.results.values():
            with self.assertRaises(TypeError):
                res.f_set(doesntwork=ChainMap({},{}))

    def test_the_insertion_made_implicetly_in_setUp(self):
        for key, val1 in self.data.items():
            res = self.results['test.res.kwargs']
            val2 = res[key]
            if isinstance(val1, dict):
                for innerkey in val1:
                    innerval1 = val1[innerkey]
                    innerval2 = val2[innerkey]
                    self.assertEqual(repr(innerval1), repr(innerval2),
                                     '%s != %s' % (str(val1),str(val2)))
            else:
                self.assertEqual(repr(val1),repr(val2), '%s != %s' % (str(val1),str(val2)))


    def test_pickling(self):
        for key, res in self.results.items():

            dump = pickle.dumps(res)

            newRes = pickle.loads(dump)

            self.results[key] = newRes

        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_meta_settings()

    def test_storage_and_loading(self):

        for key, res in self.results.items():
            store_dict = res._store()

            constructor = res.__class__


            res = constructor('')

            res._load(store_dict)

            res._rename(key)

            self.results[key] = res


        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_meta_settings()

class PickleResultTest(ResultTest):

    tags = 'unittest', 'result', 'pickle'

    def make_constructor(self):
        self.Constructor=PickleResult
        self.dynamic_imports = [PickleResult]

    def test_reject_outer_data_structure(self):
        # Since it pickles everything, it does accept all sorts of objects
        pass

    def test_meta_settings(self):
        for key, res in self.results.items():
            self.assertEqual(res.v_full_name, key)
            self.assertEqual(res.v_name, key.split('.')[-1])
            self.assertEqual(res.v_location, '.'.join(key.split('.')[0:-1]))
            self.assertEqual(res.v_protocol, self.protocols[key])

    def make_results(self):
        self.results= {}
        self.results['test.res.on_constructor']=self.Constructor('test.res.on_constructor',protocol=0,**self.data)
        self.results['test.res.args']=self.Constructor('test.res.args',protocol=1)
        self.results['test.res.kwargs']=self.Constructor('test.res.kwargs',protocol=2)

        self.protocols={'test.res.on_constructor':0,
                        'test.res.args':1,
                        'test.res.kwargs':2}

        self.results['test.res.args'].f_set(*list(self.data.values()))
        self.results['test.res.kwargs'].f_set(**self.data)

class SparseResultTest(ResultTest):

    tags = 'unittest', 'result', 'sparse'

    def make_constructor(self):
        self.Constructor=SparseResult
        self.dynamic_imports = [SparseResult]

    def setUp(self):

        if not hasattr(self,'data'):
            self.data={}

        self.data['spsparse_csc'] = spsp.csc_matrix((1000,100))
        self.data['spsparse_csc'][1,2] = 44.5

        self.data['spsparse_csr'] = spsp.csr_matrix((2222,22))
        self.data['spsparse_csr'][1,3] = 44.5

        self.data['spsparse_bsr'] = spsp.csr_matrix((111,111))
        self.data['spsparse_bsr'][3,2] = 44.5
        self.data['spsparse_bsr'] = self.data['spsparse_bsr'].tobsr()

        self.data['spsparse_dia'] = spsp.csr_matrix((111,111))
        self.data['spsparse_dia'][3,2] = 44.5
        self.data['spsparse_dia'] = self.data['spsparse_dia'].todia()

        super(SparseResultTest,self).setUp()

    def test_illegal_naming(self):
        for res in self.results.values():
            data_dict = {'val'+SparseResult.IDENTIFIER:42}
            with self.assertRaises(AttributeError):
                res.f_set(**data_dict)

    def make_results(self):
        self.results= {}
        self.results['test.res.on_constructor']=self.Constructor('test.res.on_constructor',
                                                                  protocol=0, **self.data)
        self.results['test.res.args']=self.Constructor('test.res.args')
        self.results['test.res.args'].v_protocol=1

        self.results['test.res.kwargs']=self.Constructor('test.res.kwargs', protocol=2)

        self.results['test.res.args'].f_set(*list(self.data.values()))
        self.results['test.res.kwargs'].f_set(**self.data)


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)



"""Module containing factory functions for parameter exploration"""

import logging
import sys
import itertools as itools
from collections import OrderedDict


def cartesian_product(parameter_dict, combined_parameters=()):
    """ Generates a Cartesian product of the input parameter dictionary.

    For example:

    >>> print cartesian_product({'param1':[1,2,3], 'param2':[42.0, 52.5]})
    {'param1':[1,1,2,2,3,3],'param2': [42.0,52.5,42.0,52.5,42.0,52.5]}

    :param parameter_dict:

        Dictionary containing parameter names as keys and iterables of data to explore.

    :param combined_parameters:

        Tuple of tuples. Defines the order of the parameters and parameters that are
        linked together.
        If an inner tuple contains only a single item, you can spare the
        inner tuple brackets.


        For example:

        >>> print cartesian_product( {'param1': [42.0, 52.5], 'param2':['a', 'b'], 'param3' : [1,2,3]}, ('param3',('param1', 'param2')))
        {param3':[1,1,2,2,3,3],'param1' : [42.0,52.5,42.0,52.5,42.0,52.5], 'param2':['a','b','a','b','a','b']}

    :returns: Dictionary with cartesian product lists.

    """
    if not combined_parameters:
        combined_parameters = list(parameter_dict)
    else:
        combined_parameters = list(combined_parameters)

    for idx, item in enumerate(combined_parameters):
        if isinstance(item, str):
            combined_parameters[idx] = (item,)

    iterator_list = []
    for item_tuple in combined_parameters:
        inner_iterator_list = [parameter_dict[key] for key in item_tuple]
        zipped_iterator = zip(*inner_iterator_list)
        iterator_list.append(zipped_iterator)

    result_dict = {}
    for key in parameter_dict:
        result_dict[key] = []

    cartesian_iterator = itools.product(*iterator_list)

    for cartesian_tuple in cartesian_iterator:
        for idx, item_tuple in enumerate(combined_parameters):
            for inneridx, key in enumerate(item_tuple):
                result_dict[key].append(cartesian_tuple[idx][inneridx])

    return result_dict


def find_unique_points(explored_parameters):
    """Takes a list of explored parameters and finds unique parameter combinations.

    If parameter ranges are hashable operates in O(N), otherwise O(N**2).

    :param explored_parameters:

        List of **explored** parameters

    :return:

        List of tuples, first entry being the parameter values, second entry a list
        containing the run position of the unique combination.

    """
    ranges = [param.f_get_range(copy=False) for param in explored_parameters]
    zipped_tuples = list(zip(*ranges))
    try:
        unique_elements = OrderedDict()
        for idx, val_tuple in enumerate(zipped_tuples):
            if val_tuple not in unique_elements:
                unique_elements[val_tuple] = []
            unique_elements[val_tuple].append(idx)
        return list(unique_elements.items())
    except TypeError:
        logger = logging.getLogger('pypet.find_unique')
        logger.error('Your parameter entries could not be hashed, '
                     'now I am sorting slowly in O(N**2).')
        unique_elements = []
        for idx, val_tuple in enumerate(zipped_tuples):
            matches = False
            for added_tuple, pos_list in unique_elements:
                matches = True
                for idx2, val in enumerate(added_tuple):
                    if not explored_parameters[idx2]._equal_values(val_tuple[idx2], val):
                        matches = False
                        break
                if matches:
                    pos_list.append(idx)
                    break
            if not matches:
                unique_elements.append((val_tuple, [idx]))
        return unique_elements





__author__ = 'Robert Meyer'

import numpy as np
import itertools as itools
import hashlib
from collections import deque


class Universe(object):
    """Contains everything"""
    def __contains__(self, item):
        return True


class IteratorChain(object):
    """Helper class that chains arbitrary generators and iterators and iterables.

    Preferably used over itertools.chain to avoid recursive calls.

    You can already pass some `iterators` on creation.

    """
    def __init__(self, *iterables):
        # Deque containing the iterators to come
        self._chain = deque()
        # The current iterator providing the next elements
        self._current = iter([])

        self.add(*iterables)

    def add(self, *iterables):
        """Adds `iterables` to the chain"""
        self._chain.extend(iterables)

    def next(self):
        """Returns next element from chain.

        More precisely, it returns the next element of the
        foremost iterator. If this iterator is empty it moves iteratively
        along the chain of available iterators to pick the new foremost one.

        Raises StopIteration if there are no elements left.

        """
        while True:
            # We need this loop because some iterators may already be empty.
            # We keep on popping from the left until next succeeds and as long
            # as there are iterators available
            try:
                return next(self._current)
            except StopIteration:
                try:
                    self._current = iter(self._chain.popleft())
                except IndexError:
                    # If we run out of iterators we are sure that
                    # there can be no more element
                    raise StopIteration('Reached end of iterator chain')

    def __next__(self):
        """For python 3 compatibility"""
        return self.next()

    def __iter__(self):
        while True:
            yield self.next()


class ChainMap(object):
    """Combine multiple mappings for sequential lookup.
    """

    def __init__(self, *maps):
        self._maps = maps

    def __getitem__(self, key):
        for mapping in self._maps:
            try:
                return mapping[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __len__(self):
        length = 0

        for mapping in self._maps:
            length += len(mapping)

        return length

    def __iter__(self):
        iter_list = [mapping.keys() for mapping in self._maps]
        return itools.chain(*iter_list)


class HashArray(object):
    """Hashable wrapper for numpy arrays"""

    def __init__(self, ndarray):
        """Creates a new hashable object encapsulating an ndarray.

            :param ndarray: The wrapped ndarray.

        """
        self._ndarray = ndarray

    def __eq__(self, other):

        try:
            return np.all(self._ndarray == other._ndarray)
        except AttributeError:
            return False

    def __hash__(self):
        return int(hashlib.sha1(self._ndarray.view(np.uint8)).hexdigest(), 16)


class TrajectoryMock(object):
    """Helper class that mocks properties of a trajectory.

    The full trajectory is not needed to rename a log file.
    In order to avoid copying the full trajectory during pickling
    this class is used.

    """
    def __init__(self, traj):
        self.v_environment_name = traj.v_environment_name
        self.v_name = traj.v_name
        self.v_crun_ = traj.v_crun_
        self.v_crun = traj.v_crun
        self.v_idx = traj.v_idx



__author__ = 'robert'

import random
import time
import multiprocessing as mp
import logging
import os

try:
    import scoop
    from scoop import futures
except ImportError:
    scoop = None
try:
    import zmq
except ImportError:
    zmq = None

from pypet.tests.testutils.ioutils import run_suite, make_temp_dir, remove_data, \
    get_root_logger, parse_args, unittest, get_random_port_url, errwrite
from pypet.tests.testutils.data import TrajectoryComparator
from pypet.utils.mpwrappers import LockerClient, LockerServer, TimeOutLockerServer
from pypet.pypetlogging import DisableAllLogging
from pypet.utils.helpful_functions import is_ipv6


class FaultyServer(LockerServer):
    """Simulates a server that forgets to send from time to time"""
    def __init__(self, url):
        super(FaultyServer, self).__init__(url)
        self._srep = 0

    def _pre_respond_hook(self, response):
        fail_every = 27
        load_every = 5
        respond = True
        if self._srep % fail_every == 0 and response != self.CLOSED:
            self._logger.warn('Simulating message loss; '
                              'Loosing: `%s`; '
                              'LockDB: %s' % (response, str(self._locks)))
            self._socket.close(0)
            self._context.term()
            time.sleep(2.5)
            self._context = zmq.Context()
            self._socket = self._context.socket(zmq.REP)
            self._socket.ipv6 = is_ipv6(self._url)
            self._socket.bind(self._url)
            respond = False
        elif self._srep % load_every == 0:
            self._logger.warn('Simulating heavy CPU load')
            time.sleep(0.11)

        self._srep += 1

        return respond


def run_server(server):
    # logging.basicConfig(level=1)
    server.run()


NTIMEOUTS = 3


def time_out_job(args):

    timed_out = False
    idx, lock = args
    lock.acquire()
    if idx < NTIMEOUTS:
        time.sleep(0.5)
    try:
        lock.release()
    except RuntimeError:
        if idx < NTIMEOUTS:
            timed_out = True
        else:
            raise
    return timed_out


def flush_and_sync(fh):
    fh.flush()
    try:
        os.fsync(fh.fileno())
    except OSError:
        pass


def the_job(args):
    """Simple job executed in parallel

    Just sleeps randomly and prints to console.
    Capital letter signal parallel printing

    """
    # logging.basicConfig(level=1)
    idx, lock, filename = args
    client_id = 'N/A'
    try:
        random.seed()

        sleep_time = random.uniform(0.0, 0.05)  # Random sleep time
        lock.start()
        client_id = str(lock._get_id())
        sidx = ':' + client_id + ':' + str(idx) +'\n'

        lock.acquire()
        with open(filename, mode='a') as fh:
            fh.write('PAR:__THIS__:0' + sidx)
        lock.release()
        time.sleep(sleep_time / 2.0)

        lock.acquire()
        with open(filename, mode='a') as fh:
            fh.write('PAR:__HAPPENING__:1' + sidx)
        lock.release()
        time.sleep(sleep_time * 1.5)

        lock.acquire()
        with open(filename, mode='a') as fh:
            fh.write('PAR:__PARALLEL__:2' + sidx)
        lock.release()
        time.sleep(sleep_time)

        # lock.acquire()
        # with open(filename, mode='a') as fh:
        #     fh.write('PAR:__ALL__:3' + sidx)
        # lock.release()
        # time.sleep(sleep_time / 3.0)
        #
        # lock.acquire()
        # with open(filename, mode='a') as fh:
        #     fh.write('PAR:__TIMES__:4' + sidx)
        # lock.release()
        # time.sleep(sleep_time / 1.5)


        lock.acquire()

        with open(filename, mode='a') as fh:
            fh.write('SEQ:BEGIN:0' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time / 2.0)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:This:1' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:is:2' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time * 1.5)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:a:3' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time / 3.0)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:sequential:4' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time / 1.5)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:block:5' + sidx)
            flush_and_sync(fh)
        time.sleep(sleep_time / 3.0)
        with open(filename, mode='a') as fh:
            fh.write('SEQ:END:6' + sidx)
            flush_and_sync(fh)

        lock.release()
    except:
        logging.getLogger().exception('Error in job `%d` (%s)' % (idx, client_id))


@unittest.skipIf(zmq is None, 'Cannot be run without zmq')
class TestNetLock(TrajectoryComparator):

    ITERATIONS = 33

    tags = 'unittest', 'mpwrappers', 'netlock'

    def check_file(self, filename):
        current_msg = 'END'
        current_id = -1
        current_counter = 0
        iterations = set()
        lines = []
        with open(filename) as fh:
            for line in fh:
                lines.append(line)
                split_line = line.split(':')
                if len(split_line) == 5:
                    seq, msg, counter, id_, iteration = split_line
                    if seq == 'PAR':
                        continue
                    iteration = int(iteration)
                    counter = int(counter)
                    iterations.add(iteration)
                    errstring = ('\nCurrent idx `%s` new `%s`;\n '
                               'Current msg `%s`, new `%s`;\n'
                               'Curent counter `%d`, '
                               'new `%d`;\n '
                               'Iteration %d' % (current_id, id_,
                                                 current_msg, msg,
                                                 current_counter, counter, iteration))
                    if msg == 'BEGIN':
                        self.assertEqual(current_msg, 'END', 'MSG beginning in the middle.' +
                                         errstring)

                    else:
                        self.assertEqual(current_counter, counter - 1,
                                                   'Counters not matching.' + errstring)
                        self.assertEqual(current_id, id_,
                                                   'IDs not matching.' + errstring)
                    current_counter = counter
                    current_id = id_
                    current_msg = msg
                else:
                    self.assertEqual(len(split_line), 5, 'Cannot split `%s`\n '
                                                         '---Text:---\n '
                                                         '%s' % (str(split_line),
                                                                 '\n'.join(lines)))

        self.assertEqual(len(iterations), self.ITERATIONS, '%d != %d, Iterations:\n'
                         % (len(iterations), self.ITERATIONS) +  str(iterations))
        for irun in range(self.ITERATIONS):
            self.assertIn(irun, iterations)

    def start_server(self, url, faulty=False):
        if faulty:
            ls = FaultyServer(url)
        else:
            ls = LockerServer(url)
        self.lock_process = mp.Process(target=run_server, args=(ls,))
        self.lock_process.start()

    def start_timeout_server(self, url, timeout):
        ls = TimeOutLockerServer(url, timeout)
        self.lock_process = mp.Process(target=run_server, args=(ls,))
        self.lock_process.start()

    def create_file(self, filename):
        path, file = os.path.split(filename)
        if not os.path.isdir(path):
            os.makedirs(path)
        fh = open(filename, mode='w')
        fh.close()

    def test_errors(self):
        url = get_random_port_url()
        self.start_server(url)
        ctx = zmq.Context()
        sck = ctx.socket(zmq.REQ)
        sck.ipv6 = is_ipv6(url)
        sck.connect(url)
        sck.send_string(LockerServer.UNLOCK + LockerServer.DELIMITER + 'test'
                        + LockerServer.DELIMITER + 'hi' + LockerServer.DELIMITER + '12344')
        response = sck.recv_string()
        self.assertTrue(response.startswith(LockerServer.RELEASE_ERROR))

        sck.send_string(LockerServer.LOCK + LockerServer.DELIMITER + 'test'
                        + LockerServer.DELIMITER + 'hi' + LockerServer.DELIMITER + '12344')
        response = sck.recv_string()
        self.assertEqual(response, LockerServer.GO)

        sck.send_string(LockerServer.UNLOCK + LockerServer.DELIMITER + 'test'
                        + LockerServer.DELIMITER + 'ha' + LockerServer.DELIMITER + '12344')
        response = sck.recv_string()
        self.assertTrue(response.startswith(LockerServer.RELEASE_ERROR))

        sck.send_string(LockerServer.UNLOCK + LockerServer.DELIMITER + 'test')
        response = sck.recv_string()
        self.assertTrue(response.startswith(LockerServer.MSG_ERROR))

        sck.send_string(LockerServer.LOCK + LockerServer.DELIMITER + 'test')
        response = sck.recv_string()
        self.assertTrue(response.startswith(LockerServer.MSG_ERROR))

        sck.send_string('Wooopiee!')
        response = sck.recv_string()
        self.assertTrue(response.startswith(LockerServer.MSG_ERROR))

        sck.close()

        lock = LockerClient(url)
        lock.send_done()
        self.lock_process.join()
        lock.finalize()

    def test_single_core(self):
        url = get_random_port_url()
        filename = make_temp_dir('locker_test/score.txt')
        self.create_file(filename)
        self.start_server(url)
        lock = LockerClient(url)
        iterator = [(irun, lock, filename) for irun in range(self.ITERATIONS)]
        list(map(the_job, iterator))
        lock.send_done()
        self.check_file(filename)
        self.lock_process.join()

    def test_concurrent_pool_faulty(self):
        prev = self.ITERATIONS
        self.ITERATIONS = 11
        #logging.basicConfig(level=1)
        with DisableAllLogging():
            self.test_concurrent_pool(faulty=True, filename='locker_test/faulty_pool.txt')
        self.ITERATIONS = prev

    def test_concurrent_pool(self, faulty=False, filename='locker_test/pool.txt'):
        pool = mp.Pool(5)
        url = get_random_port_url()
        filename = make_temp_dir(filename)
        self.create_file(filename)
        self.start_server(url, faulty)
        lock = LockerClient(url)
        iterator = [(irun, lock, filename) for irun in range(self.ITERATIONS)]
        pool.imap(the_job, iterator)
        pool.close()
        pool.join()
        lock.send_done()
        self.check_file(filename)
        self.lock_process.join()

    @unittest.skipIf(scoop is None, 'Can only be run with scoop')
    def test_concurrent_scoop(self):
        # test several restarts
        old_iter = self.ITERATIONS
        for jrun in range(3):
            self.ITERATIONS = 11
            url = get_random_port_url()
            filename = make_temp_dir('locker_test/scoop.txt')
            self.create_file(filename)
            self.start_server(url)
            lock = LockerClient(url)
            lock.start()
            iterator = [(irun, lock, filename) for irun in range(self.ITERATIONS)]
            list(futures.map(the_job, iterator))
            lock.send_done()
            self.check_file(filename)
            self.lock_process.join()
            # errwrite(str(irun))
        self.ITERATIONS = old_iter

    def test_timout_pool(self):
        pool = mp.Pool(5)
        url = get_random_port_url()
        self.start_timeout_server(url, 0.25)
        lock = LockerClient(url)
        lock.start()
        iterator = [(irun, lock) for irun in range(self.ITERATIONS)]
        potential_timeouts = list(pool.imap(time_out_job, iterator))
        pool.close()
        pool.join()
        all_time_outs = [x for x in potential_timeouts if x]
        self.assertGreaterEqual(len(all_time_outs), 1)
        lock.send_done()
        self.lock_process.join()

if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


__author__ = 'Robert Meyer'

import multiprocessing
# Monkey Patch from here: https://bitbucket.org/ned/coveragepy/issue/117/enable-coverage-measurement-of-code-run-by
def coverage_multiprocessing_process(): # pragma: no cover
    try:
        import coverage as _coverage
        _coverage
    except:
        return

    from coverage.collector import Collector
    from coverage import coverage
    # detect if coverage was running in forked process
    if Collector._collectors:
        original = multiprocessing.Process._bootstrap
        class Process_WithCoverage(multiprocessing.Process):
            def _bootstrap(self):
                cov = coverage(data_suffix=True,
                    omit='*/pypet/tests/*,*/shareddata.py'.split(','))

                cov.start()
                try:
                    return original(self)
                finally:
                    cov.stop()
                    cov.save()
        return Process_WithCoverage

ProcessCoverage = coverage_multiprocessing_process()
if ProcessCoverage:
    multiprocessing.Process = ProcessCoverage
    print('Added Monkey-Patch for multiprocessing and code-coverage')


import sys
import os

pypetpath=os.path.abspath(os.getcwd())
sys.path.append(pypetpath)
print('Appended path `%s`' % pypetpath)

from pypet.tests.testutils.ioutils import run_suite, discover_tests, TEST_IMPORT_ERROR, parse_args


if __name__ == '__main__':
    opt_dict = parse_args()
    tests_include = set(('TestMPImmediatePostProcLock',
                    'MultiprocFrozenPoolSortQueueTest',
                    'MultiprocFrozenPoolSortPipeTest',
                    'MultiprocLinkNoPoolLockTest',
                    'MultiprocLinkNoPoolQueueTest',
                    'MultiprocLinkQueueTest',
                    'MultiprocPoolSortLocalTest',
                    'MultiprocSCOOPSortLocalTest',
                    'MultiprocFrozenSCOOPSortNetlockTest',
                    'MultiprocFrozenSCOOPSortNetqueueTest',
                    'Brain2NetworkTest',
                    'BrainNetworkTest',
                    'CapTest'))
    pred = lambda class_name, test_name, tags: (class_name in tests_include or
                                                 'multiproc' not in tags)
    suite = discover_tests(pred)
    run_suite(suite=suite, **opt_dict)


__author__ = 'Henri Bunting'


try:
    import brian2
    from brian2 import pF, mV, defaultclock, ms, NeuronGroup, linspace, SpikeMonitor, \
        PopulationRateMonitor, StateMonitor, run, msecond, nS, nA
except ImportError:
    brian2 = None


def run_network():

    monitor_dict={}
    defaultclock.dt= 0.01*ms

    C=281*pF
    gL=30*nS
    EL=-70.6*mV
    VT=-50.4*mV
    DeltaT=2*mV
    tauw=40*ms
    a=4*nS
    b=0.08*nA
    I=8*nA
    Vcut="vm>2*mV"# practical threshold condition
    N=10

    reset = 'vm=Vr;w+=b'

    eqs="""
    dvm/dt=(gL*(EL-vm)+gL*DeltaT*exp((vm-VT)/DeltaT)+I-w)/C : volt
    dw/dt=(a*(vm-EL)-w)/tauw : amp
    Vr:volt
    """

    neuron=NeuronGroup(N,model=eqs,threshold=Vcut,reset=reset)
    neuron.vm=EL
    neuron.w=a*(neuron.vm-EL)
    neuron.Vr=linspace(-48.3*mV,-47.7*mV,N) # bifurcation parameter

    #run(25*msecond,report='text') # we discard the first spikes

    MSpike=SpikeMonitor(neuron, variables=['vm']) # record Vr and w at spike times
    MPopRate = PopulationRateMonitor(neuron)

    MMultiState = StateMonitor(neuron, ['w','vm'], record=[6,7,8,9])


    run(10*msecond,report='text')


    monitor_dict['SpikeMonitor']=MSpike
    monitor_dict['MultiState']=MMultiState
    monitor_dict['PopulationRateMonitor']=MPopRate

    return monitor_dict



if __name__ == '__main__':
    if brian2 is not None:
        run_network()


import sys
import unittest
import logging

import numpy as np
import time
import pandas as pd
from scipy import sparse as spsp
from pypet import ArrayParameter, Parameter, SparseParameter, PickleParameter, \
    PickleResult, SparseResult, ObjectTable, BaseParameter, BaseResult
from pypet.tests.testutils.ioutils import remove_data, get_root_logger
from pypet.utils.comparisons import parameters_equal, results_equal
from pypet.utils.helpful_functions import flatten_dictionary
from pypet.utils.mpwrappers import LockWrapper
from pypet.utils.siginthandling import sigint_handling

__author__ = 'Robert Meyer'


def create_param_dict(param_dict):
    '''Fills a dictionary with some parameters that can be put into a trajectory.
    '''
    param_dict['Normal'] = {}
    param_dict['Numpy'] = {}
    param_dict['Sparse'] ={}
    param_dict['Numpy_2D'] = {}
    param_dict['Numpy_3D'] = {}
    param_dict['Tuples'] ={}
    param_dict['Lists'] ={}
    param_dict['Pickle']={}

    normal_dict = param_dict['Normal']
    normal_dict['string'] = 'Im a test string!'
    normal_dict['int'] = 42
    normal_dict['long'] = 42
    normal_dict['double'] = 42.42
    normal_dict['bool'] =True
    normal_dict['trial'] = 0

    numpy_dict=param_dict['Numpy']
    numpy_dict['string'] = np.array(['Uno', 'Dos', 'Tres'])
    numpy_dict['int'] = np.array([1,2,3,4])
    numpy_dict['double'] = np.array([1.0,2.0,3.0,4.0])
    numpy_dict['bool'] = np.array([True, False, True])

    param_dict['Numpy_2D']['double'] = np.matrix([[1.0,2.0],[3.0,4.0]])
    param_dict['Numpy_3D']['double'] = np.array([[[1.0,2.0],[3.0,4.0]],[[3.0,-3.0],[42.0,41.0]]])

    spsparse_csc = spsp.lil_matrix((222,22))
    spsparse_csc[1,2] = 44.6
    spsparse_csc[1,9] = 44.5
    spsparse_csc = spsparse_csc.tocsc()

    spsparse_csr = spsp.lil_matrix((222,22))
    spsparse_csr[1,3] = 44.7
    spsparse_csr[17,17] = 44.755555
    spsparse_csr = spsparse_csr.tocsr()

    spsparse_bsr = spsp.bsr_matrix(np.matrix([[1, 1, 0, 0, 2, 2],
        [1, 1, 0, 0, 2, 2],
        [0, 0, 0, 0, 3, 3],
        [0, 0, 0, 0, 3, 3],
        [4, 4, 5, 5, 6, 6],
        [4, 4, 5, 5, 6, 6]]))

    spsparse_dia = spsp.dia_matrix(np.matrix([[1, 0, 3, 0],
        [1, 2, 0, 4],
        [0, 2, 3, 0],
        [0, 0, 3, 4]]))


    param_dict['Sparse']['bsr_mat'] = spsparse_bsr
    param_dict['Sparse']['csc_mat'] = spsparse_csc
    param_dict['Sparse']['csr_mat'] = spsparse_csr
    param_dict['Sparse']['dia_mat'] = spsparse_dia

    param_dict['Tuples']['empty'] = ()
    param_dict['Tuples']['int'] = (1,2,3)
    param_dict['Tuples']['float'] = (44.4,42.1,3.)
    param_dict['Tuples']['str'] = ('1','2wei','dr3i')

    param_dict['Lists']['lempty'] = []
    param_dict['Lists']['lint'] = [1,2,3]
    param_dict['Lists']['lfloat'] = [44.4,42.1,3.]
    param_dict['Lists']['lstr'] = ['1','2wei','dr3i']

    param_dict['Pickle']['list']= ['b','h', 53, (), 0]
    param_dict['Pickle']['list']= ['b','h', 42, (), 1]
    param_dict['Pickle']['list']= ['b',[444,43], 44, (),2]


def add_params(traj,param_dict):
    '''Adds parameters to a trajectory
    '''
    flat_dict = flatten_dictionary(param_dict,'.')

    for key, val in flat_dict.items():
        if isinstance(val, (np.ndarray, tuple)) or (isinstance(val, list) and
                                                        (len(val) < 4 or val[3] != ())):
            traj.f_add_parameter(ArrayParameter,key,val, comment='comment')
        elif isinstance(val, (str,bool,float, int)):
            traj.f_add_parameter(Parameter,key,val, comment='Im a comment!')
        elif spsp.isspmatrix(val):
            traj.f_add_parameter(SparseParameter,key,val, comment='comment').v_annotations.f_set(
                **{'Name':key,'Val' :str(val),'Favorite_Numbers:':[1,2,3],
                                 'Second_Fav':np.array([43.0,43.0])})
        elif isinstance(val,list):
            # The last item of the list `val` is an int between 0 and 2, we can use it as a
            # protocol read out to test all protocols
            traj.f_add_parameter(PickleParameter,key,val, comment='Im a comment!', protocol=val[-1])
        else:
            raise RuntimeError('You shall not pass, %s is %s!' % (str(val),str(type(val))))


    traj.f_add_derived_parameter('Another.String', 'Hi, how are you?', comment='test1')
    traj.f_add_derived_parameter('Another.StringGroup.$', 'too bad!?', comment='test2')
    traj.f_add_derived_parameter('Another.$set.$.String', 'Really?', comment='test3')
    traj.f_add_derived_parameter('Another.crun.String2', 'Really, again?', comment='test4')


    traj.f_add_result('Peter_Jackson',np.str(['is','full','of','suboptimal ideas']),
                      comment='Only my opinion bro!',)

    traj.results.f_add_leaf('Test', 42, comment='NC')
    traj.f_add_group('derived_parameters.uo', comment='Yeah, this is unsuals')
    traj.dpar.f_add_leaf('uo.adsad', 3333, comment='Yo')
    traj.derived_parameters.f_add_leaf('Test2', 42, comment='sfsdf')

    traj.par.f_add_leaf('er.Test3', 42, comment='sdfds')

    for irun in range(13):
        traj.f_add_leaf('testleaf%d' % irun, 42, comment='f')

    traj.par.f_add_group('Empty', comment='Notting!')

    traj.f_add_group('imgeneric.bitch', comment='Generic_Group')
    traj.imgeneric.f_add_leaf('gentest', 'fortytwo', comment='Oh yeah!')


def multiply(traj):
    rootlogger = get_root_logger()
    z=traj.x*traj.y
    rootlogger.info('z=x*y: '+str(z)+'='+str(traj.x)+'*'+str(traj.y))
    traj.f_add_result('z',z)
    return z


def multiply_with_graceful_exit(traj):
    z=traj.x*traj.y
    rootlogger = get_root_logger()
    rootlogger.info('z=x*y: '+str(z)+'='+str(traj.x)+'*'+str(traj.y))
    traj.f_add_result('z',z)
    if traj.v_idx > 3:
        if not sigint_handling.hit:
            sigint_handling._handle_sigint(None, None)
    return z

def multiply_with_storing(traj):
    rootlogger = get_root_logger()
    z=traj.x*traj.y
    rootlogger.info('z=x*y: '+str(z)+'='+str(traj.x)+'*'+str(traj.y))
    traj.f_add_result('z',z)
    traj.f_store()
    return z


def multiply_args(traj, arg1=0, arg2=0, arg3=0):
    rootlogger = get_root_logger()
    z=traj.x*traj.y + arg1 + arg2 + arg3
    rootlogger.info('z=x*y: '+str(z)+'='+str(traj.x)+'*'+str(traj.y) +
                    '+'+str(arg1)+'+'+str(arg2)+'+'+str(arg3))
    traj.f_add_result('z',z)
    return z


def simple_calculations(traj, arg1, simple_kwarg):

        if traj.v_idx == 0:
            # to shuffle runs
            time.sleep(0.1)

        rootlogger = get_root_logger()

        if not 'runs' in traj.res:
            traj.res.f_add_result_group('runs')

        rootlogger.info('>>>>>Starting Simple Calculations')
        my_dict = {}

        my_dict2={}
        param_dict=traj.parameters.f_to_dict(fast_access=True,short_names=False)
        for key in sorted(param_dict.keys())[0:5]:
            val = param_dict[key]
            if 'trial' in key:
                continue
            newkey = key.replace('.','_')
            my_dict[newkey] = str(val)
            my_dict2[newkey] = [str(val)+' juhu!']

        my_dict['__FLOAT'] = 44.0
        my_dict['__INT'] = 66
        my_dict['__NPINT'] = np.int_(55)
        my_dict['__INTaRRAy'] = np.array([1,2,3])
        my_dict['__FLOATaRRAy'] = np.array([1.0,2.0,41.0])
        my_dict['__FLOATaRRAy_nested'] = np.array([np.array([1.0,2.0,41.0]),np.array([1.0,2.0,41.0])])
        my_dict['__STRaRRAy'] = np.array(['sds','aea','sf'])
        my_dict['__LONG'] = 4266
        my_dict['__UNICODE'] = u'sdfdsf'
        my_dict['__BYTES'] = b'zweiundvierzig'
        my_dict['__NUMPY_UNICODE'] = np.array([u'$%&ddss'])
        my_dict['__NUMPY_BYTES'] = np.array([b'zweiundvierzig'])

        keys = sorted(to_dict_wo_config(traj).keys())
        for idx,key in enumerate(keys[0:5]):
            keys[idx] = key.replace('.', '_')

        listy=traj.f_add_result_group('List', comment='Im a result group')
        traj.f_add_result_group('Iwiiremainempty.yo', comment='Empty group!')
        traj.Iwiiremainempty.f_store_child('yo')

        traj.Iwiiremainempty.f_add_link('kkk',listy )
        listy.f_add_link('hhh', traj.Iwiiremainempty)

        if not traj.Iwiiremainempty.kkk.v_full_name == traj.List.v_full_name:
            raise RuntimeError()

        if not traj.Iwiiremainempty.kkk.v_full_name == traj.List.hhh.kkk.v_full_name:
            raise RuntimeError()

        traj.f_add_result('runs.' + traj.v_crun + '.ggg', 5555, comment='ladida')
        traj.res.runs.f_add_result(traj.v_crun + '.ggjg', 5555, comment='didili')
        traj.res.runs.f_add_result('hhg', 5555, comment='jjjj')

        traj.res.f_add_result(name='lll', comment='duh', data=444)

        x = traj.res.f_add_result(name='nested', comment='duh')

        x['nested0.nested1.nested2.nested3'] =  44

        traj.res.f_add_result(name='test.$set.$', comment='duh', data=444)

        try:
            traj.f_add_config('teeeeest', 12)
            raise RuntimeError()
        except TypeError:
            pass

        # if not traj.f_contains('results.runs.' + traj.v_crun + '.ggjg', shortcuts=False):
        #     raise RuntimeError()
        # if not traj.f_contains('results.runs.' + traj.v_crun + '.ggg', shortcuts=False):
        #     raise RuntimeError()
        if not traj.f_contains('results.runs.' + traj.v_crun + '.hhg', shortcuts=False):
            raise RuntimeError()

        traj.f_add_result('List.Of.Keys', dict1=my_dict, dict2=my_dict2, comment='Test')
        traj.List.f_store_child('Of', recursive=True)
        traj.f_add_result('DictsNFrame', keys=keys, comment='A dict!')
        traj.f_add_result('ResMatrix',np.array([1.2,2.3]), comment='ResMatrix')

        traj.f_add_result('empty.stuff', (), [], {}, np.array([]), comment='empty stuff')
        #traj.f_add_derived_parameter('All.To.String', str(traj.f_to_dict(fast_access=True,short_names=False)))

        myframe = pd.DataFrame(data ={'TC1':[1,2,3],'TC2':['Waaa','',''],'TC3':[1.2,42.2,77]})

        myseries = myframe['TC1']

        traj.f_add_result('myseries', myseries, comment='dd')
        traj.f_store_item('myseries')

        traj.f_get('DictsNFrame').f_set(myframe)

        traj.f_add_result('IStore.SimpleThings',1.0,3,np.float32(5.0),
                          'Iamstring',
                          (1,2,3), [4,5,6],
                          zwei=2).v_comment='test'
        traj.f_add_derived_parameter('super.mega',33, comment='It is huuuuge!')
        traj.super.f_set_annotations(AgainATestAnnotations='I am a string!111elf')

        traj.f_add_result(PickleResult,'pickling.result.proto1', my_dict2, protocol=1, comment='p1')
        traj.f_add_result(PickleResult,'pickling.result.proto2', my_dict2, protocol=2, comment='p2')
        traj.f_add_result(PickleResult,'pickling.result.proto0', my_dict2, protocol=0, comment='p0')

        traj.f_add_result(SparseResult, 'sparse.csc', traj.csc_mat, 42).v_comment='sdsa'
        traj.f_add_result(SparseResult, 'sparse.bsr', traj.bsr_mat, 52).v_comment='sdsa'
        traj.f_add_result(SparseResult, 'sparse.csr', traj.csr_mat, 62).v_comment='sdsa'
        traj.f_add_result(SparseResult, 'sparse.dia', traj.dia_mat, 72).v_comment='sdsa'

        traj.sparse.v_comment = 'I contain sparse data!'

        myobjtab = ObjectTable(data={'strings':['a','abc','qwertt'], 'ints':[1,2,3]})

        traj.f_add_result('object.table', myobjtab, comment='k').v_annotations.f_set(test=42)
        traj.object.f_set_annotations(test2=42.42)

        traj.f_add_result('$.here', 77, comment='huhu')
        traj.f_add_result('tor.hot.$', dollah=77, comment='duh!')
        traj.f_add_result('tor.hot.rrr.$.j', 77, comment='duh!')
        traj.f_add_result('tor.hot.rrr.crun.jjj', 777, comment='duh**2!')

        #traj.f_add_result('PickleTerror', result_type=PickleResult, test=traj.SimpleThings)
        rootlogger.info('<<<<<<Finished Simple Calculations')

        # let's see if the traj can also always be returned
        if isinstance(traj.v_storage_service, LockWrapper):
            traj.v_storage_service.pickle_lock = False
        return 42, traj


def to_dict_wo_config(traj):
        res_dict={}

        for child_name in traj._children:

            child = traj._children[child_name]
            if child_name == 'config':
                continue

            if child.v_is_leaf:
                res_dict[child_name] = child
            else:
                res_dict.update(child.f_to_dict(fast_access=False))

        return res_dict


class TrajectoryComparator(unittest.TestCase):

    def are_results_in_order(self, results):
        self.assertGreater(len(results), 0)
        sorted_res = sorted(results, key=lambda key: key[0])
        self.assertEqual(results, sorted_res)

    def clear_handlers(self):
        """Deletes all handlers and closes all log-files"""
        root = logging.getLogger()
        for logger in list(root.manager.loggerDict.values()) + [root]:
            if hasattr(logger, 'handlers'):
                handlers = logger.handlers
                for handler in handlers:
                    if hasattr(handler, 'flush'):
                        handler.flush()
                    if hasattr(handler, 'close'):
                        handler.close()
                logger.handlers = []
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def tearDown(self):
        if hasattr(self, 'env') and hasattr(self.env, 'f_disable_logging'):
            self.env.f_disable_logging()
        self.clear_handlers()
        remove_data()

    def compare_trajectories(self,traj1,traj2):

        trajlength = len(traj1)

        if traj1.f_get_run_information(0)['completed']:

            # We need this fix in case the trajectory was never run
            # thus it does not contain any run groups but still has a length
            if 'results.runs' in traj1:
                rungroups = traj1.results.runs.f_children()
            else:
                rungroups = 1

            self.assertEqual(trajlength, rungroups, 'len of traj1 is %d, rungroups %d' % (trajlength, rungroups))

        old_items = to_dict_wo_config(traj1)
        new_items = to_dict_wo_config(traj2)

        self.assertEqual(len(traj1),len(traj2), 'Length unequal %d != %d.' % (len(traj1), len(traj2)))

        self.assertEqual(len(old_items),len(new_items))
        for key,item in new_items.items():
            old_item = old_items[key]

            if isinstance(item, BaseParameter):
                self.assertTrue(parameters_equal(item,old_item),
                                'For key %s: %s not equal to %s' % (key,str(old_item),str(item)))
            elif isinstance(item,BaseResult):
                self.assertTrue(results_equal(item, old_item),
                                'For key %s: %s not equal to %s' % (key, str(old_item),str(item)))
            else:
                raise RuntimeError('You shall not pass')


            self.assertTrue(str(item.v_annotations)==str(old_item.v_annotations),'%s != %s' %
                        (item.v_annotations.f_ann_to_str(),old_item.v_annotations.f_ann_to_str()))

        # Check the annotations
        for node in traj1.f_iter_nodes(recursive=True):

            if node.v_run_branch == traj1.f_wildcard('$', 0) or node.v_run_branch == 'trajectory':
                if node.v_comment != '' and node.v_full_name in traj2:
                    second_comment = traj2.f_get(node.v_full_name).v_comment
                    self.assertEqual(node.v_comment, second_comment, '%s != %s, for %s' %
                                                (node.v_comment, second_comment, node.v_full_name))

            if not node.v_annotations.f_is_empty():
                second_anns = traj2.f_get(node.v_full_name).v_annotations
                self.assertTrue(str(node.v_annotations) == str(second_anns))



"""Helper module that allows parsing of `.ini` files"""

__author__ = 'Robert Meyer'

import functools
import ast
import os
import configparser as cp

from pypet.pypetlogging import use_simple_logging


def parse_config(init_func):
    """Decorator wrapping the environment to use a config file"""
    @functools.wraps(init_func)
    def new_func(env, *args, **kwargs):
        config_interpreter = ConfigInterpreter(kwargs)
        # Pass the config data to the kwargs
        new_kwargs = config_interpreter.interpret()
        init_func(env, *args, **new_kwargs)
        # Add parameters and config data from the `.ini` file
        config_interpreter.add_parameters(env.traj)
    return new_func


class ConfigInterpreter(object):
    """ Helper class that parses an `.ini` file and passes the data to an environment."""
    def __init__(self, kwargs):
        self.kwargs = kwargs
        self.config_file = kwargs.pop('config', None)
        self.parser = None
        if self.config_file:
            if isinstance(self.config_file, str):
                # Check if the config file exists
                if not os.path.isfile(self.config_file):
                    raise ValueError('`%s` does not exist.' % self.config_file)
                # If yes parse it with a config parser
                self.parser = cp.ConfigParser()
                self.parser.read(self.config_file)
            elif isinstance(self.config_file, cp.RawConfigParser):
                # Already instantiaded parsers are also accepted
                self.parser = self.config_file
            else:
                raise RuntimeError('Your config file/parser format `%s` '
                                   'is not understood.' % str(self.config_file))

    def _collect_section(self, section):
        """Collects all settings within a section"""
        kwargs = {}
        try:
            if self.parser.has_section(section):
                options = self.parser.options(section)
                for option in options:
                    str_val = self.parser.get(section, option)
                    val = ast.literal_eval(str_val)
                    kwargs[option] = val
            return kwargs
        except:
            raise  # You can set a break point here for debugging!

    def _collect_config(self):
        """Collects all info from three sections"""
        kwargs = {}
        sections = ('storage_service', 'trajectory', 'environment')
        for section in sections:
            kwargs.update(self._collect_section(section))
        return kwargs

    def interpret(self):
        """Copies parsed arguments into the kwargs passed to the environment"""
        if self.config_file:
            new_kwargs = self._collect_config()
            for key in new_kwargs:
                # Already specified kwargs take precedence over the ini file
                if key not in self.kwargs:
                    self.kwargs[key] = new_kwargs[key]
            if not use_simple_logging(self.kwargs) and 'log_config' not in self.kwargs:
                self.kwargs['log_config'] = self.config_file
        return self.kwargs

    def add_parameters(self, traj):
        """Adds parameters and config from the `.ini` file to the trajectory"""
        if self.config_file:
            parameters = self._collect_section('parameters')
            for name in parameters:
                value = parameters[name]
                if not isinstance(value, tuple):
                    value = (value,)
                traj.f_add_parameter(name, *value)
            config = self._collect_section('config')
            for name in config:
                value = config[name]
                if not isinstance(value, tuple):
                    value = (value,)
                traj.f_add_config(name, *value)



__author__ = 'Robert Meyer'

import pypet.pypetconstants as pypetconstants
from pypet.tests.integration.link_test import LinkEnvironmentTest
from pypet.tests.testutils.ioutils import parse_args, run_suite
try:
    import psutil
except ImportError:
    psutil = None


class MultiprocLinkQueueTest(LinkEnvironmentTest):

    tags = 'integration', 'hdf5', 'environment', 'multiproc', 'queue', 'pool', 'links'

    def set_mode(self):
        LinkEnvironmentTest.set_mode(self)
        self.mode = pypetconstants.WRAP_MODE_QUEUE
        self.multiproc = True
        self.log_stdout = True
        if psutil is not None:
            self.ncores = 0
        else:
            self.ncores = 3
        self.use_pool=True


class MultiprocLinkLocalTest(LinkEnvironmentTest):

    tags = 'integration', 'hdf5', 'environment', 'multiproc', 'local', 'pool', 'links'

    def set_mode(self):
        LinkEnvironmentTest.set_mode(self)
        self.mode = pypetconstants.WRAP_MODE_LOCAL
        self.multiproc = True
        self.log_stdout = True
        self.use_pool=True


class MultiprocLinkLockTest(LinkEnvironmentTest):

    tags = 'integration', 'hdf5', 'environment', 'multiproc', 'lock', 'pool', 'links'

    # def test_run(self):
    #     super(MultiprocLockTest, self).test_run()

    def set_mode(self):
        LinkEnvironmentTest.set_mode(self)
        self.mode = pypetconstants.WRAP_MODE_LOCK
        self.multiproc = True
        self.ncores = 4
        self.use_pool=True


class MultiprocLinkNoPoolQueueTest(LinkEnvironmentTest):

    tags = 'integration', 'hdf5', 'environment', 'multiproc', 'queue', 'nopool', 'links'

    def set_mode(self):
        LinkEnvironmentTest.set_mode(self)
        self.mode = pypetconstants.WRAP_MODE_QUEUE
        self.multiproc = True
        self.ncores = 3
        self.use_pool=False


class MultiprocLinkNoPoolLockTest(LinkEnvironmentTest):

    tags = 'integration', 'hdf5', 'environment', 'multiproc', 'lock', 'nopool', 'links'

    def set_mode(self):
        LinkEnvironmentTest.set_mode(self)
        self.mode = pypetconstants.WRAP_MODE_LOCK
        self.multiproc = True
        self.ncores = 2
        self.use_pool=False


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


"""
New module allowing to share data across processes in a multiprocessing environment.

Already functional but not yet documented or thoroughly tested.

Use at your own risk!

"""

__author__ = 'Robert Meyer'

import warnings

import pandas as pd
import numpy as np

import pypet.pypetconstants as pypetconstants
from pypet.pypetlogging import HasLogger
from pypet.utils.decorators import with_open_store, prefix_naming
from pypet.parameter import ObjectTable, Result
from pypet.naturalnaming import KnowsTrajectory


@prefix_naming
class StorageContextManager(HasLogger):
    """Context manager that can be used to open a storage file with multiprocessing.

    For example:

    .. code-block:: python

        with StorageContextManager(traj) as scm:
            # This is a block with open storage,
            # this block is also multiprocessing safe,
            # i.e. everything happens sequentially in here
            do_stuff_to_open_storage_sequentially()

    """
    def __init__(self, trajectory):
        self._traj = trajectory

    def __enter__(self):
        self.open_store()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close_store()
        except Exception as exc:
            self._logger.error('Could not close file because of `%s`' % repr(exc))
            if exc_type is None:
                raise
            else:
                return False

    def close_store(self):
        """Closes store manually not needed if used with `with`"""
        service = self._traj.v_storage_service
        if not service.is_open:
            raise RuntimeError('The storage service is not open, '
                               'please open via `f_open_storage`.')
        service.store(pypetconstants.CLOSE_FILE, None)

    def open_store(self):
        """Opens store manually not needed if used with `with`"""
        service = self._traj.v_storage_service
        if service.is_open:
            raise RuntimeError('Your service is already open, there is no need to re-open it.')
        service.store(pypetconstants.OPEN_FILE, None,
                      trajectory_name=self._traj.v_name)

    def flush_store(self):
        """Flushes data to the storage can be called at any time when storage is open."""
        service = self._traj.v_storage_service
        if not service.is_open:
            raise RuntimeError('The storage service is not open, '
                               'please open via `f_open_storage`.')
        service.store(pypetconstants.FLUSH, None)


def make_ordinary_result(result, key, trajectory=None, reload=True):
    """Turns a given shared data item into a an ordinary one.

    :param result: Result container with shared data
    :param key: The name of the shared data
    :param trajectory:

        The trajectory, only needed if shared data has
        no access to the trajectory, yet.

    :param reload: If data should be reloaded after conversion

    :return: The result


    """
    shared_data = result.f_get(key)
    if trajectory is not None:
        shared_data.traj = trajectory
    shared_data._request_data('make_ordinary')
    result.f_remove(key)
    if reload:
        trajectory.f_load_item(result, load_data=pypetconstants.OVERWRITE_DATA)
    return result


def make_shared_result(result, key, trajectory, new_class=None):
    """Turns an ordinary data item into a shared one.

    Removes the old result from the trajectory and replaces it.
    Empties the given result.

    :param result: The result containing ordinary data
    :param key: Name of ordinary data item
    :param trajectory: Trajectory container
    :param new_class:

        Class of new shared data item.
        Leave `None` for automatic detection.

    :return: The `result`

    """
    data = result.f_get(key)
    if new_class is None:
        if isinstance(data, ObjectTable):
            new_class = SharedTable
        elif isinstance(data, pd.DataFrame):
            new_class = SharedPandasFrame
        elif isinstance(data, (tuple, list)):
            new_class = SharedArray
        elif isinstance(data, (np.ndarray, np.matrix)):
            new_class = SharedCArray
        else:
            raise RuntimeError('Your data `%s` is not understood.' % key)
    shared_data = new_class(result.f_translate_key(key), result, trajectory=trajectory)
    result[key] = shared_data
    shared_data._request_data('make_shared')
    return result


class SharedData(HasLogger):
    """ Piece of data that can be shared among many processes.

    Could be a PyTables table or array or pandas DataFrame.
    The data can be accessed and modified by all processes.

    Can be added as follows:

        >>> res = traj.f_add_result(SharedResult, 'shared_result')
        >>> my_array = SharedArray(name='array', parent=res, trajectory=traj, add_to_parent=True)

    Next, the shared array can be accessed via:

        >>> my_array = traj.shared_result.array

    So far the shared array is empty and we need to create shared data on the disk.
    In order to do this we call

        >>> my_array.create_shared_data(obj=np.zeros((100,100)))

    Thus, within the HDF5 file, there exist now a 100 by 100 array that can be
    accessed and modified by any process.

    For example:

    .. code-block:: python

        with StorageContextManager(traj) as scm:
            my_array = traj.shared_result
            my_array[42, 42] = 1001.array

    Now the value 1001 can be accessed from all processes at
    position ``[42, 42]``.

    :param name:

        Name of shared data, must be the same name as in the parent node.

    :param parent:

        *pypet* result that will contain the shared data item.

    :param trajectory:

        Trajectory container

    :param add_to_parent:

        If shared data should automatically be added to the parent
        with the given `name`.

        >>> my_array = SharedArray(name='array', parent=res, trajectory=traj, add_to_parent=True)

        is equivalent to

        >>> my_array = SharedArray(name='array', parent=res, trajectory=traj, add_to_parent=False)
        >>> res['array'] = my_array

    """

    FLAG = None

    def __init__(self, name=None, parent=None, trajectory=None, add_to_parent=False):
        self._set_logger()
        self.name = name
        self.parent = parent
        self.traj = trajectory
        if add_to_parent:
            self.parent[name] = self

    def _check_state(self):
        if self.traj is None:
            raise TypeError('Please pass the trajectory to the shared data, via '
                            '`shared_data.traj = trajectory`, otherwise you cannot '
                            'access the data.')
        if self.parent is None:
            raise TypeError('Please pass the parent to the shared data, via'
                            '`shared_data.parent = parent`, otherwise you cannot '
                            'access the data.')
        if self.name is None:
            raise TypeError('Please pass the name of the shared data, via'
                            '`shared_data.name = name`, otherwise you cannot '
                            'access the data.')

        if self.name not in self.parent or self.parent.f_get(self.name) is not self:
            raise TypeError('Please make sure that the SharedData is added to the '
                            'parent with the name of the SharedData.')

    def _store_parent(self):
        self._check_state()
        if not self.parent._stored:
            self.traj.f_store_item(self.parent)

    def create_shared_data(self, **kwargs):
        """Creates shared data on disk with a StorageService on disk.

        Needs to be called before shared data can be used later on.

        Actual arguments of ``kwargs`` depend on the type of data to be
        created. For instance, creating an array one can use the keyword
        ``obj`` to pass a numpy array (``obj=np.zeros((10,20,30))``).
        Whereas for a PyTables table may need a description dictionary
        (``description={'column_1': pt.StringCol(2, pos=0),'column_2': pt.FloatCol( pos=1)}``)
        Refer to the PyTables documentation on how to create tables.

        """
        if 'flag' not in kwargs:
            kwargs['flag'] = self.FLAG
        if 'data' in kwargs:
            kwargs['obj'] = kwargs.pop('data')
        if 'trajectory' in kwargs:
            self.traj = kwargs.pop('trajectory')
        if 'traj' in kwargs:
            self.traj = kwargs.pop('traj')
        if 'name' in kwargs:
            self.name = kwargs.pop['name']
        if 'parent' in kwargs:
            self.parent = kwargs.pop('parent')
            if self.name is not None:
                self.parent[self.name] = self
        return self._request_data('create_shared_data', kwargs=kwargs)

    def _request_data(self, request, args=None, kwargs=None):
        """Interface with the underlying storage.

        Passes request to the StorageService that performs the appropriate action.
        For example, given a shared table ``t``.
        ``t.remove_row(4)`` is parsed into ``request='remove_row', args=(4,)`` and
        passed onto the storage service. In case of the HDF5StorageService,
        this is again translated back into ``hdf5_table_node.remove_row(4)``.

        """
        return self._storage_service.store(pypetconstants.ACCESS_DATA,
                                           self.parent.v_full_name,
                                           self.name,
                                           request, args, kwargs,
                                           trajectory_name=self.traj.v_name)

    def get_data_node(self):
        """Returns the actula node of the underlying data.

        In case one uses HDF5 this will be the HDF5 leaf node.

        """
        if not self._storage_service.is_open:
            warnings.warn('You requesting the data item but your store is not open, '
                          'the item itself will be closed, too!',
                          category=RuntimeWarning)
        return self._request_data('__thenode__')

    @property
    def _storage_service(self):
        self._store_parent()
        return self.traj.v_storage_service


class SharedArray(SharedData):

    FLAG = pypetconstants.ARRAY

    @property
    def rowsize(self):
        return self._request_data('rowsize')

    @property
    def nrows(self):
        return self._request_data('nrows')

    @property
    def atom(self):
        return self._request_data('atom')

    def get_enum(self):
        return self._request_data('get_enum', args=())

    def read(self, start=None, stop=None, step=None, out=None):
        kwargs = dict(start=start, stop=stop, step=step, out=out)
        return self._request_data('read', kwargs=kwargs)

    def __iter__(self):
        if self._storage_service.is_open:
            return self.iterrows()
        else:
            return iter(self.read())

    def __getitem__(self, item):
        return self._request_data('__getitem__', args=(item,))

    def __setitem__(self, key, value):
        return self._request_data('__setitem__', args=(key, value))

    def __len__(self):
        return self.nrows

    @with_open_store
    def iterrows(self, start=None, stop=None, step=None):
        return self._request_data('iterrows', kwargs=dict(start=start, stop=stop, step=step))

    @with_open_store
    def next(self):
        return self._request_data('next', args=())


class SharedCArray(SharedArray):

    FLAG = pypetconstants.CARRAY


class SharedEArray(SharedArray):

    FLAG = pypetconstants.EARRAY

    def append(self, sequence):
        return self._request_data('append', args=(sequence,))


class SharedVLArray(SharedEArray):

    FLAG = pypetconstants.VLARRAY


class SharedTable(SharedData):

    FLAG = pypetconstants.TABLE

    @property
    def coldescrs(self):
        return self._request_data('coldescrs')

    @property
    def coldtypes(self):
        return self._request_data('coldtypes')

    @property
    def colindexed(self):
        return self._request_data('colindexed')

    @property
    def colnames(self):
        return self._request_data('colnames')

    @property
    def colpathnames(self):
        return self._request_data('colpathnames')

    @property
    def coltypes(self):
        return self._request_data('coltypes')

    @property
    def description(self):
        return self._request_data('description')

    @property
    def indexed(self):
        return self._request_data('indexed')

    @property
    def extdim(self):
        return self._request_data('extdim')

    @property
    def nrows(self):
        return self._request_data('nrows')

    @property
    def rowsize(self):
        return self._request_data('rowsize')

    @property
    def autoindex(self):
        return self._request_data('autoindex')

    @autoindex.setter
    def autoindex(self, value):
        self._request_data('__setattr__', args=('autoindex', value))

    @property
    def indexedcolpathnames(self):
        return self._request_data('indexedcolpathnames')

    def col(self, name):
        return self._request_data('col', args=(name,))

    def read(self, start=None, stop=None, step=None, field=None, out=None):
        kwargs = dict(start=start, stop=stop, step=step, field=field, out=out)
        return self._request_data('read', kwargs=kwargs)

    def read_coordinates(self, coords, field=None):
        return self._request_data('read_coordinates', args=(coords,), kwargs=dict(filed=field))

    def read_sorted(self, sortby, checkCSI=False, field=None, start=None, stop=None, step=None):
        kwargs = dict(checkCSI=checkCSI, field=field, start=start, stop=stop, step=step)
        return self._request_data('read_sorted', args=(sortby,), kwargs=kwargs)

    def __iter__(self):
        if self._storage_service.is_open:
            return self._request_data('__iter__', args=())
        else:
            return iter(self.read())

    def __getitem__(self, key):
        return self._request_data('__getitem__', args=(key,))

    def __len__(self):
        return self.nrows

    def append(self, rows):
        return self._request_data('append', args=(rows,))

    # def append_row(self, row):
    #     return self.append([row])

    def modify_column(self, start=None, stop=None, step=None, column=None, colname=None):
        kwargs = dict(start=start, stop=stop, step=step, column=column, colname=colname)
        return self._request_data('modify_column', kwargs=kwargs)

    def modify_columns(self, start=None, stop=None, step=None, column=None, colname=None):
        kwargs = dict(start=start, stop=stop, step=step, column=column, colname=colname)
        return self._request_data('modify_columns', kwargs=kwargs)

    def modify_coordinates(self, coords, rows):
        self._request_data('modify_coordinates', args=(coords, rows))

    def modify_rows(self, start=None, stop=None, step=None, rows=None):
        kwargs = dict(start=start, stop=stop, step=step, rows=rows)
        self._request_data('modify_rows', kwargs=kwargs)

    def remove_rows(self, start=None, stop=None, step=None):
        kwargs = dict(start=start, stop=stop, step=step)
        self._request_data('remove_rows', kwargs=kwargs)

    def remove_row(self, n):
        self._request_data('remove_row', args=(n,))

    def __setitem__(self, key, value):
        return self._request_data('__setitem__', args=(key, value))

    def get_where_list(self, condition, condvars=None,
                         sort=False, start=None, stop=None, step=None):
        kwargs = dict(condvars=condvars, sort=sort, start=start, stop=stop, step=step)
        return self._request_data('get_where_list', args=(condition,), kwargs=kwargs)

    def read_where(self, condition, condvars=None, field=None,
                     start=None, stop=None, step=None):
        kwargs = dict(condvars=condvars, field=field, start=start, stop=stop, step=step)
        return self._request_data('read_where', args=(condition,), kwargs=kwargs)

    def will_query_use_indexing(self, condition, condvars=None):
        return self._request_data('will_query_use_indexing', args=(condition,),
                                  kwargs=dict(condvars=condvars))

    def get_enum(self, colname):
        return self._request_data('get_enum', args=(colname,))

    def reindex(self, colname=None):
        return self._request_data('reindex', kwargs=dict(colname=colname,
                                                         _col_func=colname is not None))

    def reindex_dirty(self, colname=None):
        return self._request_data('reindex_dirty', kwargs=dict(colname=colname,
                                                                _col_func=colname is not None))

    def remove_index(self, colname):
        return self._request_data('remove_index', args=(colname,),
                                  kwargs=dict(_col_func=True))

    def create_index(self, colname, optlevel=6,
                       kind='medium', filters=None,
                       tmp_dir=None, _blocksizes=None,
                       _testmode=False, _verbose=False):
        kwargs = dict(optlevel=optlevel, kind=kind, filters=filters, tmp_dir=tmp_dir,
                      _blocksizes=_blocksizes, _testmode=_testmode, _verbose=_verbose,
                      _col_func=True, colname=colname)
        return self._request_data('create_index', kwargs=kwargs)

    def create_csindex(self, colname, filters=None, tmp_dir=None,
                         _blocksizes=None, _testmode=False, _verbose=False):
        kwargs = dict(filters=filters, tmp_dir=tmp_dir, _blocksizes=_blocksizes,
                      _testmode=_testmode, _verbose=_verbose,
                      _col_func=True, colname=colname)

        return self._request_data('create_csindex', kwargs=kwargs)

    @property
    @with_open_store
    def colindexes(self):
        return self._request_data('colindexes')

    @property
    @with_open_store
    def cols(self):
        return self._request_data('cols')

    @property
    @with_open_store
    def row(self):
        return self._request_data('row')

    @with_open_store
    def iterrows(self, start=None, stop=None, step=None):
        return self._request_data('iterrows', kwargs=dict(start=start, stop=stop, step=step))

    @with_open_store
    def itersequence(self, sequence):
        return self._request_data('itersequence', args=(sequence,))

    @with_open_store
    def itersorted(self, sortby, checkCSI=False, start=None, stop=None, step=None):
        kwargs = dict(checkCSI=checkCSI, start=start, stop=stop, step=step)
        return self._request_data('itersorted', args=(sortby,), kwargs=kwargs)

    @with_open_store
    def where(self, condition, condvars=None, start=None, stop=None, step=None):
        kwargs = dict(condvars=condvars, start=start, stop=stop, step=step)
        return self._request_data('where', args=(condition,), kwargs=kwargs)

    @with_open_store
    def append_where(self, dstTable, condition, condvars=None, start=None, stop=None, step=None):
        kwargs = dict(condvars=condvars, start=start, stop=stop, step=step)
        return self._request_data('append_where', args=(dstTable, condition), kwargs=kwargs)

    @with_open_store
    def flush(self):
        return self._request_data('flush', args=())

    @with_open_store
    def flush_rows_to_index(self, _lastrow=True):
        return self._request_data('flush_rows_to_index', kwargs=dict(_lastrow=_lastrow))


class SharedPandasFrame(SharedData):

    FLAG = pypetconstants.FRAME

    def create_shared_data(self, **kwargs):
        if 'format' not in kwargs:
            kwargs['format'] = 'table'
        return super(SharedPandasFrame, self).create_shared_data(**kwargs)

    def put(self, data=None, format='table', append=False, **kwargs):
        # data = self._extract_data(data)
        kwargs['format'] = format
        kwargs['append'] = append
        kwargs['obj'] = data
        return self._request_data('pandas_put', kwargs=kwargs)

    def append(self, data=None, **kwargs):
        # data = self._extract_data(data)
        return self.put(data, format='table', append=True, **kwargs)

    def select(self, where=None, start = None, stop=None, columns=None,
               chunksize=None, **kwargs):
        kwargs['where'] = where,
        kwargs['start'] = start
        kwargs['stop'] = stop
        kwargs['columns'] = columns
        kwargs['chunksize'] = chunksize
        if not self._storage_service.is_open:
            kwargs['iterator'] = False
        return self._request_data('pandas_select', kwargs=kwargs)
        # return self.data

    def read(self):
        return self._request_data('pandas_get', args=())
        # return self.data


FLAG_CLASS_MAPPING = {
    pypetconstants.ARRAY: SharedArray,
    pypetconstants.CARRAY: SharedCArray,
    pypetconstants.EARRAY: SharedEArray,
    pypetconstants.VLARRAY: SharedVLArray,
    pypetconstants.TABLE: SharedTable,
    pypetconstants.FRAME: SharedPandasFrame,
}


class SharedResult(Result, KnowsTrajectory):
    """Behaves exactly like the normal `Result` but accepts `SharedData` subclasses as data."""

    __slots__ = ('_traj',)

    SUPPORTED_DATA = set(FLAG_CLASS_MAPPING.values())

    def __init__(self, full_name, trajectory, *args, **kwargs):
        self._traj = trajectory
        super(SharedResult, self).__init__(full_name, *args, **kwargs)

    @property
    def v_traj(self):
        return self._traj

    def _supports(self, item):
        """Checks if outer data structure is supported."""
        result = super(SharedResult, self)._supports(item)
        result = result or type(item) in SharedResult.SUPPORTED_DATA
        return result

    def _pass_to_shared(self, name, item):
        try:
            item.traj = self.v_traj
            item.name = name
            item.parent = self
        except AttributeError:
            pass

    def f_set_single(self, name, item):
        super(SharedResult, self).f_set_single(name, item)
        self._pass_to_shared(name, item)

    def _load(self, load_dict):
        super(SharedResult, self)._load(load_dict)
        for key in self:
            item = self[key]
            self._pass_to_shared(key, item)

    def create_shared_data(self, name=None, **kwargs):
        """Calls the corresponding function of the shared data item"""
        if name is None:
            item = self.f_get()
        else:
            item = self.f_get(name)
        return item.create_shared_data(**kwargs)



__author__ = 'Robert Meyer'

import logging
import itertools as itools
import os
import platform
import sys
import unittest

from pypet import Result
from pypet.tests.testutils.data import TrajectoryComparator
from pypet.tests.testutils.ioutils import parse_args, run_suite, make_temp_dir,\
    make_trajectory_name,  get_log_config, get_log_path
from pypet import Environment, rename_log_file, Trajectory, Parameter
import pypet.pypetlogging


class LogWhenStored(Result):

    def __init__(self, full_name, *args, **kwargs):
        self._level = kwargs['level']
        super(LogWhenStored, self).__init__(full_name, *args, **kwargs)

    def _store(self):
        self._logger.log(self._level, 'STORE_Test! in parameter')
        return super(LogWhenStored, self)._store()


def add_result(traj, level=logging.ERROR):
    traj.f_ares(LogWhenStored, 'logging.test',
                42, level=level, comment='STORE_Test!')


def log_error(traj):
    add_result(traj)
    logging.getLogger('pypet.only.ERROR').error('Test of Error Logging during run!')


def log_wo_error_levels(traj):
    add_result(traj, level=logging.INFO)
    logging.getLogger('pypet.DEBUG').debug('DEBUG_Test!')
    logging.getLogger('pypet.INFO').info('INFO_Test!')
    logging.getLogger('pypet.WARNING').warning('WARNING_Test!')


def log_all_levels(traj):
    add_result(traj)
    logging.getLogger('pypet.DEBUG').debug('DEBUG_Test!')
    logging.getLogger('pypet.INFO').info('INFO_Test!')
    logging.getLogger('pypet.WARNING').warning('WARNING_Test!')
    logging.getLogger('pypet.ERROR').error('ERROR_Test!')
    logging.getLogger('pypet.CRITICAL').critical('CRITICAL_Test!')


class Dummy(object):
    pass


class LoggingTest(TrajectoryComparator):

    tags = 'integration', 'environment', 'logging'

    def setUp(self):
        root = logging.getLogger()
        for logger in itools.chain(root.manager.loggerDict.values(), [root]):
            if hasattr(logger, 'handlers'):
                for handler in logger.handlers:
                    if hasattr(handler, 'flush'):
                        handler.flush()
                    if hasattr(handler, 'close'):
                        handler.close()
                logger.handlers = []
            if hasattr(logger, 'setLevel'):
                logger.setLevel(logging.NOTSET)
        self.set_mode()

    def tearDown(self):
        super(LoggingTest, self).tearDown()

    def set_mode(self):
        self.mode = Dummy()
        self.mode.wrap_mode = 'LOCK'
        self.mode.multiproc = False
        self.mode.ncores = 1
        self.mode.use_pool=True
        self.mode.pandas_format='fixed'
        self.mode.pandas_append=False
        self.mode.complib = 'blosc'
        self.mode.complevel=9
        self.mode.shuffle=True
        self.mode.fletcher32 = False
        self.mode.encoding = 'utf8'
        self.mode.log_stdout=False
        self.mode.log_config=get_log_config()


    def make_env(self, **kwargs):

        self.mode.__dict__.update(kwargs)
        filename = 'log_testing.hdf5'
        self.filename = make_temp_dir(filename)
        self.traj_name = make_trajectory_name(self)
        self.env = Environment(trajectory=self.traj_name,
                               filename=self.filename, **self.mode.__dict__)
        self.traj = self.env.v_traj


    def add_params(self, traj):

        traj.par.p1 = Parameter('', 42, 'Hey')
        traj.f_apar('g1.p2', 145, comment='Test')


    def explore(self, traj):
        traj.f_explore({'p1': range(7)})

    # @unittest.skipIf(platform.system() == 'Windows', 'Log file creation might fail under windows.')
    def test_logfile_creation_normal(self):
        # if not self.multiproc:
        #     return
        self.make_env()
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_wo_error_levels)
        self.env.f_disable_logging()

        traj = self.env.v_traj

        log_path = get_log_path(traj)

        if self.mode.multiproc:
            if self.mode.use_pool:
                length = self.mode.ncores * 2
            else:
                length = 2 * len(traj)
            if self.mode.wrap_mode == 'LOCK':
                length += 2
            elif self.mode.wrap_mode == 'QUEUE':
                length += 4
            else:
                raise RuntimeError('You shall not pass!')
        else:
            length = 2


        file_list = [file for file in os.listdir(log_path)]

        self.assertEqual(len(file_list), length) # assert that there are as many
        # files as runs plus main.txt and errors and warnings
        total_error_count = 0
        total_store_count = 0
        total_info_count = 0
        total_retry_count = 0
        for file in file_list:
            with open(os.path.join(log_path, file), mode='r') as fh:
                text = fh.read()
            if len(text) == 0:
                continue
            count = text.count('INFO_Test!')
            total_info_count += count
            error_count = text.count('ERROR_Test!')
            total_error_count += error_count
            store_count = text.count('STORE_Test!')
            total_store_count += store_count
            retry_count = text.count('Retry')
            total_retry_count += retry_count
            if 'LOG.txt' == file:
                if self.mode.multiproc:
                    self.assertEqual(count,0)
                    self.assertEqual(store_count, 0)
                else:
                    self.assertEqual(count, len(traj))
                    self.assertEqual(store_count, len(traj))
            elif 'ERROR' in file:
                full_path = os.path.join(log_path, file)
                filesize = os.path.getsize(full_path)
                with open(full_path) as fh:
                    text = fh.read()
                if 'Retry' not in text:
                    self.assertEqual(filesize, 0)
            elif 'Queue' in file:
                self.assertEqual(store_count, len(traj))
            elif 'LOG' in file:
                if self.mode.multiproc and self.mode.use_pool:
                    self.assertGreaterEqual(count, 0, '%d < 1 for file %s' % (count, file))
                else:
                    self.assertEqual(count, 1)
                    if self.mode.wrap_mode == 'QUEUE':
                        self.assertEqual(store_count, 0)
                    else:
                        self.assertEqual(store_count, 1)
            else:
                self.assertTrue(False, 'There`s a file in the log folder that does not '
                                       'belong there: %s' % str(file))
        self.assertEqual(total_store_count, len(traj))
        self.assertEqual(total_error_count, 0)
        self.assertEqual(total_info_count, len(traj))
        self.assertLess(total_retry_count, len(traj))

    def test_throw_error_when_specifying_config_and_old_method(self):
        with self.assertRaises(ValueError):
            self.make_env(log_config=None, logger_names='test')

    def test_disable(self):
        # if not self.multiproc:
        #     return
        self.make_env(log_config=None)
        traj = self.env.v_traj

        log_path = get_log_path(traj)

        self.assertFalse(os.path.isdir(log_path))
        self.assertTrue(self.env._logging_manager._sp_config is None)
        self.assertTrue(self.env._logging_manager._mp_config is None)
        self.assertTrue(self.env._logging_manager.log_config is None)

        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)

        self.assertFalse(os.path.isdir(log_path))
        self.assertTrue(self.env._logging_manager._sp_config is None)
        self.assertTrue(self.env._logging_manager._mp_config is None)
        self.assertTrue(self.env._logging_manager.log_config is None)

        self.env.f_disable_logging()
        # pypet_path = os.path.abspath(os.path.dirname(pypet.pypetlogging))
        # init_path = os.path.join(pypet_path, 'logging')
        # log_config = os.path.join(init_path, 'default.ini')


    # @unittest.skipIf(platform.system() == 'Windows', 'Log file creation might fail under windows.')
    def test_logfile_creation_with_errors(self):
         # if not self.multiproc:
        #     return
        self.make_env()
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)
        if self.mode.multiproc:
            logging.getLogger('pypet.test').error('ttt')
        self.env.f_disable_logging()

        traj = self.env.v_traj
        log_path = get_log_path(traj)

        if self.mode.multiproc:
            if self.mode.use_pool:
                length = self.mode.ncores * 2
            else:
                length = 2 * len(traj)
            if self.mode.wrap_mode == 'LOCK':
                length += 2
            elif self.mode.wrap_mode == 'QUEUE':
                length += + 4
            else:
                raise RuntimeError('You shall not pass!')
        else:
            length = 2

        file_list = [file for file in os.listdir(log_path)]

        self.assertEqual(len(file_list), length) # assert that there are as many
        # files as runs plus main.txt and errors and warnings

        total_error_count = 0
        total_store_count = 0
        total_info_count = 0
        total_retry_count = 0
        for file in file_list:
            with open(os.path.join(log_path, file), mode='r') as fh:
                text = fh.read()
            if len(text) == 0:
                continue
            count = text.count('INFO_Test!')
            total_info_count += count
            error_count = text.count('ERROR_Test!')
            total_error_count += error_count
            store_count = text.count('STORE_Test!')
            total_store_count += store_count
            retry_count = text.count('Retry')
            total_retry_count += retry_count
            if 'LOG.txt' == file:
                if self.mode.multiproc:
                    self.assertEqual(count,0)
                    self.assertEqual(store_count, 0)
                else:
                    self.assertEqual(count, len(traj))
                    self.assertEqual(store_count, len(traj))
            elif 'ERROR.txt' == file:
                self.assertEqual(count, 0)
                if self.mode.multiproc:
                    self.assertEqual(error_count,0)
                    self.assertEqual(store_count, 0)
                else:
                    self.assertEqual(error_count, len(traj))
                    self.assertEqual(store_count, len(traj))

            elif 'Queue' in file and 'ERROR' in file:
                self.assertEqual(store_count, len(traj))
            elif 'Queue' in file and 'LOG' in file:
                self.assertEqual(store_count, len(traj))
            elif 'LOG' in file:
                if self.mode.multiproc and self.mode.use_pool:
                    self.assertGreaterEqual(count, 0)
                    self.assertGreaterEqual(error_count, 0)
                else:
                    self.assertEqual(count, 1)
                    self.assertEqual(error_count, 1)
                    if self.mode.wrap_mode == 'QUEUE':
                        self.assertEqual(store_count, 0)
                    else:
                        self.assertEqual(store_count, 1)
            elif 'ERROR' in file:
                if self.mode.multiproc and self.mode.use_pool:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(error_count, 1)
                else:
                    self.assertEqual(count, 0)
                    self.assertEqual(error_count, 1)
                    if self.mode.wrap_mode == 'QUEUE':
                        self.assertEqual(store_count, 0)
                    else:
                        self.assertEqual(store_count, 1)
            else:
                self.assertTrue(False, 'There`s a file in the log folder that does not '
                                       'belong there: %s' % str(file))
        self.assertEqual(total_store_count, 2*len(traj))
        self.assertEqual(total_error_count, 2*len(traj))
        self.assertEqual(total_info_count, len(traj))
        self.assertLess(total_retry_count, len(traj))

    def test_file_renaming(self):
        traj_name = 'test'
        traj = Trajectory('test', add_time=False)
        traj.f_add_parameter('x', 42)
        traj.f_explore({'x': [1,2,3]})
        rename_string = '$traj_$set_$run'
        solution_1 = 'test_run_set_ALL_run_ALL'
        solution_2 = 'test_run_set_00000_run_00000000'
        renaming_1 = rename_log_file(rename_string, traj)
        self.assertEqual(renaming_1, solution_1)
        traj.v_idx = 0
        renaming_2 = rename_log_file(rename_string, traj)
        self.assertEqual(renaming_2, solution_2)


    # @unittest.skipIf(platform.system() == 'Windows', 'Log file creation might fail under windows.')
    def test_logfile_old_way_creation_with_errors(self):
         # if not self.multiproc:
        #     return
        del self.mode.__dict__['log_config']
        self.make_env(logger_names = ('','pypet'), log_level=logging.ERROR,
                      log_folder=make_temp_dir('logs'))
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)
        if self.mode.multiproc:
            logging.getLogger('pypet.test').error('ttt')
        self.env.f_disable_logging()

        traj = self.env.v_traj
        log_path = get_log_path(traj)

        if self.mode.multiproc:
            if self.mode.use_pool:
                length = self.mode.ncores * 2
            else:
                length = 2 * len(traj)
            if self.mode.wrap_mode == 'LOCK':
                length += 2
            elif self.mode.wrap_mode == 'QUEUE':
                length += 4
            else:
                raise RuntimeError('You shall not pass!')
        else:
            length = 2

        file_list = [file for file in os.listdir(log_path)]

        self.assertEqual(len(file_list), length) # assert that there are as many
        # files as runs plus main.txt and errors and warnings

        total_error_count = 0
        total_store_count = 0
        total_info_count = 0
        total_retry_count = 0

        for file in file_list:
            with open(os.path.join(log_path, file), mode='r') as fh:
                text = fh.read()
            if len(text) == 0:
                continue
            count = text.count('INFO_Test!')
            total_info_count += count
            error_count = text.count('ERROR_Test!')
            total_error_count += error_count
            store_count = text.count('STORE_Test!')
            total_store_count += store_count
            retry_count = text.count('Retry')
            total_retry_count += retry_count
            if 'LOG.txt' == file:
                if self.mode.multiproc:
                    self.assertEqual(count,0)
                    self.assertEqual(store_count, 0)
                else:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(store_count, len(traj))
            elif 'ERROR.txt' == file:
                self.assertEqual(count, 0)
                if self.mode.multiproc:
                    self.assertEqual(error_count,0)
                    self.assertEqual(store_count, 0)
                else:
                    self.assertGreaterEqual(error_count, len(traj))
                    self.assertGreaterEqual(store_count, len(traj))

            elif 'Queue' in file and 'ERROR' in file:
                self.assertGreaterEqual(store_count, len(traj))
            elif 'Queue' in file and 'LOG' in file:
                self.assertGreaterEqual(store_count, len(traj))
            elif 'LOG' in file:
                if self.mode.multiproc and self.mode.use_pool:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(error_count, 0)
                else:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(error_count, 1)
                    if self.mode.wrap_mode == 'QUEUE':
                        self.assertEqual(store_count, 0)
                    else:
                        self.assertGreaterEqual(store_count, 1)
            elif 'ERROR' in file:
                if self.mode.multiproc and self.mode.use_pool:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(error_count, 0)
                else:
                    self.assertEqual(count, 0)
                    self.assertGreaterEqual(error_count, 1)
                    if self.mode.wrap_mode == 'QUEUE':
                        self.assertEqual(store_count, 0)
                    else:
                        self.assertGreaterEqual(store_count, 1)
            else:
                self.assertTrue(False, 'There`s a file in the log folder that does not '
                                       'belong there: %s' % str(file))
        self.assertGreaterEqual(total_store_count, 2*len(traj))
        self.assertGreaterEqual(total_error_count, 2*len(traj))
        self.assertEqual(total_info_count, 0)
        self.assertLess(total_retry_count, len(traj))

    # @unittest.skipIf(platform.system() == 'Windows', 'Log file creation might fail under windows.')
    def test_logfile_old_way_disabling_mp_log(self):
         # if not self.multiproc:
        #     return
        del self.mode.__dict__['log_config']
        self.make_env(logger_names = ('','pypet'), log_level=logging.ERROR,
                      log_folder=make_temp_dir('logs'), log_multiproc=False)
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)
        if self.mode.multiproc:
            logging.getLogger('pypet.test').error('ttt')
        self.env.f_disable_logging()

        traj = self.env.v_traj
        log_path = get_log_path(traj)

        # if self.mode.multiproc:
        length = 2

        file_list = [file for file in os.listdir(log_path)]

        self.assertEqual(len(file_list), length) # assert that there are as many
        # files as runs plus main.txt and errors and warnings

        # total_error_count = 0
        # total_store_count = 0
        # total_info_count = 0
        # total_retry_count = 0
        #
        # for file in file_list:
        #     with open(os.path.join(log_path, file), mode='r') as fh:
        #         text = fh.read()
        #     count = text.count('INFO_Test!')
        #     total_info_count += count
        #     error_count = text.count('ERROR_Test!')
        #     total_error_count += error_count
        #     store_count = text.count('STORE_Test!')
        #     total_store_count += store_count
        #     retry_count = text.count('Retry')
        #     total_retry_count += retry_count
        #     if 'LOG.txt' == file:
        #         if self.mode.multiproc:
        #             self.assertEqual(count,0)
        #             self.assertEqual(store_count, 0)
        #         else:
        #             self.assertEqual(count, 0)
        #             self.assertGreaterEqual(store_count, len(traj))
        #     elif 'ERROR.txt' == file:
        #         self.assertEqual(count, 0)
        #         if self.mode.multiproc:
        #             self.assertEqual(error_count,0)
        #             self.assertEqual(store_count, 0)
        #         else:
        #             self.assertGreaterEqual(error_count, len(traj))
        #             self.assertGreaterEqual(store_count, len(traj))
        #
        #     elif 'Queue' in file and 'ERROR' in file:
        #         self.assertGreaterEqual(store_count, len(traj))
        #     elif 'Queue' in file and 'LOG' in file:
        #         self.assertGreaterEqual(store_count, len(traj))
        #     elif 'LOG' in file:
        #         if self.mode.multiproc and self.mode.use_pool:
        #             self.assertEqual(count, 0)
        #             self.assertGreaterEqual(error_count, 0)
        #         else:
        #             self.assertEqual(count, 0)
        #             self.assertGreaterEqual(error_count, 1)
        #             if self.mode.wrap_mode == 'QUEUE':
        #                 self.assertEqual(store_count, 0)
        #             else:
        #                 self.assertGreaterEqual(store_count, 1)
        #     elif 'ERROR' in file:
        #         if self.mode.multiproc and self.mode.use_pool:
        #             self.assertEqual(count, 0)
        #             self.assertGreaterEqual(error_count, 0)
        #         else:
        #             self.assertEqual(count, 0)
        #             self.assertGreaterEqual(error_count, 1)
        #             if self.mode.wrap_mode == 'QUEUE':
        #                 self.assertEqual(store_count, 0)
        #             else:
        #                 self.assertGreaterEqual(store_count, 1)
        #     else:
        #         self.assertTrue(False, 'There`s a file in the log folder that does not '
        #                                'belong there: %s' % str(file))
        # self.assertGreaterEqual(total_store_count, 2*len(traj))
        # self.assertGreaterEqual(total_error_count, 2*len(traj))
        # self.assertEqual(total_info_count, 0)
        # self.assertLess(total_retry_count, len(traj))

    # @unittest.skipIf(platform.system() == 'Windows', 'Log file creation might fail under windows.')
    def test_logging_stdout(self):
        filename = 'teststdoutlog.hdf5'
        filename = make_temp_dir(filename)
        folder = make_temp_dir('logs')
        env = Environment(trajectory=make_trajectory_name(self),
                          filename=filename, log_config=get_log_config(),
                          # log_levels=logging.CRITICAL, # needed for the test
                          log_stdout=('STDOUT', 50), #log_folder=folder
                          )

        env.f_run(log_error)
        traj = env.v_traj
        path = get_log_path(traj)

        mainstr = 'sTdOuTLoGGinG'
        print(mainstr)
        env.f_disable_logging()

        mainfilename = os.path.join(path, 'LOG.txt')
        with open(mainfilename, mode='r') as mainf:
            full_text = mainf.read()

        self.assertTrue(mainstr in full_text)
        self.assertTrue('4444444' not in full_text)
        self.assertTrue('DEBUG' not in full_text)


    def test_logging_show_progress(self):
        self.make_env(log_config=get_log_config(),
                      # log_folder=make_temp_dir('logs'),
                      # log_levels=logging.ERROR,
                      report_progress=(3, 'progress', 40))
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)
        self.env.f_disable_logging()

        traj = self.env.v_traj

        path = get_log_path(traj)
        mainfilename = os.path.join(path, 'LOG.txt')
        with open(mainfilename, mode='r') as mainf:
            full_text = mainf.read()

        progress = 'PROGRESS: Finished'
        self.assertTrue(progress in full_text)
        bar = '[=='
        self.assertTrue(bar in full_text)


    def test_logging_show_progress_print(self):
        self.make_env(log_config=get_log_config(), log_stdout=('prostdout', 50),
                      report_progress=(3, 'print'))
        self.add_params(self.traj)
        self.explore(self.traj)

        self.env.f_run(log_all_levels)
        self.env.f_disable_logging()

        path = get_log_path(self.env.v_traj)
        mainfilename = os.path.join(path, 'LOG.txt')
        with open(mainfilename, mode='r') as mainf:
            full_text = mainf.read()

        progress = 'prostdout CRITICAL PROGRESS: Finished'
        self.assertTrue(progress in full_text)
        bar = '[=='
        self.assertIn(bar, full_text)


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


__author__ = ['Henri Bunting', 'Robert Meyer']


from pypet.brian2.parameter import Brian2Parameter, Brian2Result, Brian2MonitorResult
from pypet.brian2.network import NetworkManager, NetworkRunner, NetworkComponent, NetworkAnalyser


__all__ = [
    Brian2Parameter.__name__,
    Brian2Result.__name__,
    Brian2MonitorResult.__name__,
    NetworkManager.__name__,
    NetworkRunner.__name__,
    NetworkComponent.__name__,
    NetworkAnalyser.__name__
]